# Generated by Django 5.2.1 on 2025-05-14 20:33

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('tag', '0003_alter_tag_sort_order'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='taggeditem',
            name='tag_taggedi_content_cd6f88_idx',
        ),
        migrations.RenameIndex(
            model_name='tag',
            new_name='tag_created_at_ix',
            old_name='tag_created_at_idx',
        ),
        migrations.RenameIndex(
            model_name='tag',
            new_name='tag_updated_at_ix',
            old_name='tag_updated_at_idx',
        ),
        migrations.RenameIndex(
            model_name='tag',
            new_name='tag_sort_order_ix',
            old_name='tag_sort_order_idx',
        ),
        migrations.RenameIndex(
            model_name='tag',
            new_name='tag_active_ix',
            old_name='tag_tag_active_5d272a_btree',
        ),
        migrations.AddIndex(
            model_name='tag',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['id'], name='tag_id_ix'),
        ),
        migrations.AddIndex(
            model_name='taggeditem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['created_at'], name='taggeditem_created_at_ix'),
        ),
        migrations.AddIndex(
            model_name='taggeditem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['updated_at'], name='taggeditem_updated_at_ix'),
        ),
        migrations.AddIndex(
            model_name='taggeditem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['content_type', 'object_id'], name='tagged_item_content_obj_ix'),
        ),
        migrations.AddIndex(
            model_name='taggeditem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['tag'], name='tagged_item_tag_ix'),
        ),
        migrations.AddIndex(
            model_name='taggeditem',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['object_id'], name='tagged_item_object_id_ix'),
        ),
    ]



import importlib

import factory
from django.apps import apps
from faker import Faker

from cart.models import CartItem

fake = Faker()


def get_or_create_product():
    if apps.get_model("product", "Product").objects.exists():
        return (
            apps.get_model("product", "Product").objects.order_by("?").first()
        )
    else:
        product_factory_module = importlib.import_module(
            "product.factories.product"
        )
        product_factory_class = product_factory_module.ProductFactory
        return product_factory_class.create()


def get_or_create_cart(is_guest=False):
    from cart.factories.cart import CartFactory  # noqa: PLC0415

    if apps.get_model("cart", "Cart").objects.exists():
        if is_guest:
            cart = (
                apps.get_model("cart", "Cart")
                .objects.filter(user__isnull=True)
                .order_by("?")
                .first()
            )
            if cart:
                return cart
        else:
            cart = (
                apps.get_model("cart", "Cart")
                .objects.filter(user__isnull=False)
                .order_by("?")
                .first()
            )
            if cart:
                return cart

        return apps.get_model("cart", "Cart").objects.order_by("?").first()
    elif is_guest:
        return CartFactory(is_guest=True)
    else:
        return CartFactory()


class CartItemFactory(factory.django.DjangoModelFactory):
    cart = factory.LazyFunction(get_or_create_cart)
    product = factory.LazyFunction(get_or_create_product)
    quantity = factory.Faker("random_int", min=1, max=10)

    class Meta:
        model = CartItem
        django_get_or_create = ("cart", "product")

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        is_guest = kwargs.pop("is_guest", False)
        if is_guest and "cart" not in kwargs:
            kwargs["cart"] = get_or_create_cart(is_guest=True)
        return super()._create(model_class, *args, **kwargs)



# Generated by Django 5.2.1 on 2025-05-14 20:33

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pay_way', '0006_alter_payway_sort_order'),
    ]

    operations = [
        migrations.RenameIndex(
            model_name='payway',
            new_name='payway_created_at_ix',
            old_name='payway_created_at_idx',
        ),
        migrations.RenameIndex(
            model_name='payway',
            new_name='payway_updated_at_ix',
            old_name='payway_updated_at_idx',
        ),
        migrations.RenameIndex(
            model_name='payway',
            new_name='payway_sort_order_ix',
            old_name='payway_sort_order_idx',
        ),
        migrations.RenameIndex(
            model_name='payway',
            new_name='pay_way_active_ix',
            old_name='pay_way_pay_active_1c54dd_btree',
        ),
        migrations.AddIndex(
            model_name='payway',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['cost'], name='pay_way_cost_ix'),
        ),
        migrations.AddIndex(
            model_name='payway',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['free_for_order_amount'], name='pay_way_free_order_ix'),
        ),
    ]



from django.urls import reverse
from rest_framework.test import APITestCase

from blog.factories.tag import BlogTagFactory


class BlogTagFilterTest(APITestCase):
    def setUp(self):
        self.tag1 = BlogTagFactory()
        self.tag2 = BlogTagFactory()

    def test_active_filter(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"active": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"active": "false"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_name_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"name__icontains": "test"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"name__icontains": "tag"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"name__istartswith": "t"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_post_count_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"has_posts": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"min_posts": 1})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"max_posts": 10})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_post_relationship_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"post__author": 1})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"post__category": 1})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"post__is_published": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_engagement_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"has_engagement": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"min_engagement": 1})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_timestamp_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(
            url, {"created_after": "2024-01-01T00:00:00Z"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(
            url, {"created_before": "2025-12-31T23:59:59Z"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_uuid_and_sort_order_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"uuid": str(self.tag1.uuid)})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"sort_order__gte": 1})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_camel_case_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"isActive": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"hasPosts": "false"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_complex_filter_combinations(self):
        url = reverse("blog-tag-list")

        response = self.client.get(
            url, {"active": "true", "name__icontains": "T", "has_posts": "true"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"min_posts": 1, "active": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_special_filters(self):
        url = reverse("blog-tag-list")

        response = self.client.get(url, {"most_used": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"unused": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_filter_with_ordering(self):
        url = reverse("blog-tag-list")

        response = self.client.get(
            url, {"active": "true", "ordering": "sort_order"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)



# Generated by Django 5.0.4 on 2024-04-11 17:27
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("product", "0010_alter_productcategory_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="productcategory",
            options={
                "ordering": ["sort_order"],
                "verbose_name": "Product Category",
                "verbose_name_plural": "Product Categories",
            },
        ),
        migrations.AlterModelOptions(
            name="productimage",
            options={
                "ordering": ["sort_order"],
                "verbose_name": "Product Image",
                "verbose_name_plural": "Product Images",
            },
        ),
    ]



# Generated by Django 5.2.3 on 2025-06-21 14:42

import django.contrib.postgres.indexes
import djmoney.models.fields
from decimal import Decimal
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pay_way', '0008_payway_configuration_payway_is_online_payment_and_more'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='payway',
            name='pay_way_free_order_ix',
        ),
        migrations.RenameField(
            model_name='payway',
            old_name='free_for_order_amount_currency',
            new_name='free_threshold_currency',
        ),
        migrations.RemoveField(
            model_name='payway',
            name='free_for_order_amount',
        ),
        migrations.AddField(
            model_name='payway',
            name='free_threshold',
            field=djmoney.models.fields.MoneyField(decimal_places=2, default=Decimal('0'), help_text='Order amount above which this payment method becomes free', max_digits=11, verbose_name='Free Threshold'),
        ),
        migrations.AddIndex(
            model_name='payway',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['free_threshold'], name='pay_way_free_threshold_ix'),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-11 17:27
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("region", "0003_alter_region_options"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="region",
            options={
                "ordering": ["sort_order"],
                "verbose_name": "Region",
                "verbose_name_plural": "Regions",
            },
        ),
    ]



from django.test import TestCase

from blog.factories.author import BlogAuthorFactory
from blog.factories.comment import BlogCommentFactory
from blog.factories.post import BlogPostFactory
from blog.factories.tag import BlogTagFactory
from blog.managers.post import BlogPostQuerySet
from blog.models.post import BlogPost
from user.factories.account import UserAccountFactory


class TestBlogPostManager(TestCase):
    def setUp(self):
        self.author1 = BlogAuthorFactory(user=UserAccountFactory())
        self.author2 = BlogAuthorFactory(user=UserAccountFactory())

        self.post1 = BlogPostFactory(author=self.author1)
        self.post2 = BlogPostFactory(author=self.author2)
        self.post3 = BlogPostFactory(author=self.author1)

        self.comment1 = BlogCommentFactory(
            post=self.post1, user=UserAccountFactory(), approved=True
        )
        self.comment2 = BlogCommentFactory(
            post=self.post1, user=UserAccountFactory(), approved=True
        )
        self.comment3 = BlogCommentFactory(
            post=self.post1,
            user=UserAccountFactory(),
            approved=False,
        )
        self.comment4 = BlogCommentFactory(
            post=self.post2, user=UserAccountFactory(), approved=True
        )

        self.tag1 = BlogTagFactory(active=True)
        self.tag2 = BlogTagFactory(active=True)
        self.tag3 = BlogTagFactory(active=False)

        self.post1.tags.add(self.tag1, self.tag2, self.tag3)
        self.post2.tags.add(self.tag1)

    def test_with_likes_count_manager_method(self):
        posts = BlogPost.objects.with_likes_count_annotation()

        assert posts.count() == 3

        for post in posts:
            assert hasattr(post, "likes_count_annotation")
            assert isinstance(post.likes_count_annotation, int)

    def test_with_comments_count_manager_method(self):
        posts = BlogPost.objects.with_comments_count_annotation()

        assert posts.count() == 3

        post1_annotated = posts.get(id=self.post1.id)
        post2_annotated = posts.get(id=self.post2.id)
        post3_annotated = posts.get(id=self.post3.id)

        assert post1_annotated.comments_count_annotation == 2
        assert post2_annotated.comments_count_annotation == 1
        assert post3_annotated.comments_count_annotation == 0

    def test_with_tags_count_manager_method(self):
        posts = BlogPost.objects.with_tags_count_annotation()

        assert posts.count() == 3

        post1_annotated = posts.get(id=self.post1.id)
        post2_annotated = posts.get(id=self.post2.id)
        post3_annotated = posts.get(id=self.post3.id)

        assert post1_annotated.tags_count_annotation == 2
        assert post2_annotated.tags_count_annotation == 1
        assert post3_annotated.tags_count_annotation == 0

    def test_manager_returns_correct_queryset_type(self):
        queryset = BlogPost.objects.with_likes_count_annotation()

        assert isinstance(queryset, BlogPostQuerySet)

    def test_model_properties_vs_annotations(self):
        posts = BlogPost.objects.with_comments_count_annotation()

        for post in posts:
            assert post.comments_count == post.comments_count_annotation


class TestBlogPostQuerySet(TestCase):
    def setUp(self):
        self.author1 = BlogAuthorFactory(user=UserAccountFactory())
        self.author2 = BlogAuthorFactory(user=UserAccountFactory())

        self.post1 = BlogPostFactory(author=self.author1)
        self.post2 = BlogPostFactory(author=self.author2)
        self.post3 = BlogPostFactory(author=self.author1)

        self.user1 = UserAccountFactory()
        self.user2 = UserAccountFactory()
        self.user3 = UserAccountFactory()

        self.post1.likes.add(self.user1, self.user2)
        self.post2.likes.add(self.user1)

        BlogCommentFactory(post=self.post1, user=self.user1, approved=True)
        BlogCommentFactory(post=self.post1, user=self.user2, approved=True)
        BlogCommentFactory(post=self.post1, user=self.user3, approved=False)
        BlogCommentFactory(post=self.post2, user=self.user1, approved=True)

        self.active_tag1 = BlogTagFactory(active=True)
        self.active_tag2 = BlogTagFactory(active=True)
        self.inactive_tag = BlogTagFactory(active=False)

        self.post1.tags.add(
            self.active_tag1, self.active_tag2, self.inactive_tag
        )
        self.post2.tags.add(self.active_tag1)
        self.post3.tags.add(self.inactive_tag)

    def test_with_likes_count_queryset_method(self):
        queryset = BlogPost.objects.filter().with_likes_count_annotation()

        assert queryset.count() == 3

        post1_annotated = queryset.get(id=self.post1.id)
        post2_annotated = queryset.get(id=self.post2.id)
        post3_annotated = queryset.get(id=self.post3.id)

        assert post1_annotated.likes_count_annotation == 2
        assert post2_annotated.likes_count_annotation == 1
        assert post3_annotated.likes_count_annotation == 0

    def test_with_comments_count_queryset_method(self):
        queryset = BlogPost.objects.filter().with_comments_count_annotation()

        assert queryset.count() == 3

        post1_annotated = queryset.get(id=self.post1.id)
        post2_annotated = queryset.get(id=self.post2.id)
        post3_annotated = queryset.get(id=self.post3.id)

        assert post1_annotated.comments_count_annotation == 2
        assert post2_annotated.comments_count_annotation == 1
        assert post3_annotated.comments_count_annotation == 0

    def test_with_tags_count_queryset_method(self):
        queryset = BlogPost.objects.filter().with_tags_count_annotation()

        assert queryset.count() == 3

        post1_annotated = queryset.get(id=self.post1.id)
        post2_annotated = queryset.get(id=self.post2.id)
        post3_annotated = queryset.get(id=self.post3.id)

        assert post1_annotated.tags_count_annotation == 2
        assert post2_annotated.tags_count_annotation == 1
        assert post3_annotated.tags_count_annotation == 0

    def test_queryset_chaining(self):
        queryset = (
            BlogPost.objects.filter(author=self.author1)
            .with_likes_count_annotation()
            .with_comments_count_annotation()
            .with_tags_count_annotation()
        )

        assert queryset.count() == 2

        for post in queryset:
            assert hasattr(post, "likes_count_annotation")
            assert hasattr(post, "comments_count_annotation")
            assert hasattr(post, "tags_count_annotation")

    def test_queryset_with_ordering(self):
        queryset = BlogPost.objects.with_likes_count_annotation().order_by(
            "-likes_count_annotation"
        )

        posts = list(queryset)

        assert posts[0].id == self.post1.id
        assert posts[1].id == self.post2.id
        assert posts[2].id == self.post3.id

    def test_queryset_with_filtering(self):
        queryset = BlogPost.objects.with_comments_count_annotation().filter(
            comments_count_annotation__gt=0
        )

        assert queryset.count() == 2
        assert self.post1 in queryset
        assert self.post2 in queryset
        assert self.post3 not in queryset

    def test_distinct_behavior_with_annotations(self):
        self.post1.likes.add(UserAccountFactory())
        BlogCommentFactory(
            post=self.post1, user=UserAccountFactory(), approved=True
        )

        queryset = (
            BlogPost.objects.with_likes_count_annotation()
            .with_comments_count_annotation()
            .with_tags_count_annotation()
        )

        post_ids = list(queryset.values_list("id", flat=True))
        assert len(post_ids) == len(set(post_ids))

    def test_queryset_returns_correct_type(self):
        queryset = BlogPost.objects.filter().with_likes_count_annotation()

        assert isinstance(queryset, BlogPostQuerySet)

        chained = queryset.with_comments_count_annotation()
        assert isinstance(chained, BlogPostQuerySet)

    def test_approved_comments_filtering(self):
        queryset = BlogPost.objects.with_comments_count_annotation()
        post1_annotated = queryset.get(id=self.post1.id)

        assert post1_annotated.comments_count_annotation == 2

        assert (
            post1_annotated.comments_count_annotation
            == self.post1.comments_count
        )
        assert self.post1.all_comments_count == 3

    def test_active_tags_filtering(self):
        queryset = BlogPost.objects.with_tags_count_annotation()
        post1_annotated = queryset.get(id=self.post1.id)

        assert post1_annotated.tags_count_annotation == 2

        assert post1_annotated.tags_count_annotation == self.post1.tags_count


class TestBlogPostManagerEdgeCases(TestCase):
    def test_empty_queryset_behavior(self):
        assert BlogPost.objects.with_likes_count_annotation().count() == 0
        assert BlogPost.objects.with_comments_count_annotation().count() == 0
        assert BlogPost.objects.with_tags_count_annotation().count() == 0

    def test_posts_without_relations(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        likes_queryset = BlogPost.objects.with_likes_count_annotation()
        comments_queryset = BlogPost.objects.with_comments_count_annotation()
        tags_queryset = BlogPost.objects.with_tags_count_annotation()

        likes_post = likes_queryset.get(id=post.id)
        comments_post = comments_queryset.get(id=post.id)
        tags_post = tags_queryset.get(id=post.id)

        assert likes_post.likes_count_annotation == 0
        assert comments_post.comments_count_annotation == 0
        assert tags_post.tags_count_annotation == 0

    def test_posts_with_inactive_tags_only(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        inactive_tag1 = BlogTagFactory(active=False)
        inactive_tag2 = BlogTagFactory(active=False)
        post.tags.add(inactive_tag1, inactive_tag2)

        annotated_posts = BlogPost.objects.with_tags_count_annotation()
        annotated_post = annotated_posts.get(id=post.id)

        assert annotated_post.tags_count_annotation == 0

    def test_posts_with_unapproved_comments_only(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=False)
        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=False)

        annotated_posts = BlogPost.objects.with_comments_count_annotation()
        annotated_post = annotated_posts.get(id=post.id)

        assert annotated_post.comments_count_annotation == 0
        assert post.all_comments_count == 2

    def test_large_dataset_performance(self):
        author = BlogAuthorFactory(user=UserAccountFactory())

        posts = []
        for i in range(20):
            post = BlogPostFactory(author=author)
            posts.append(post)

            for j in range(i % 5):
                user = UserAccountFactory()
                post.likes.add(user)
                BlogCommentFactory(post=post, user=user, approved=j % 2 == 0)

                if j < 3:
                    tag = BlogTagFactory(active=True)
                    post.tags.add(tag)

        likes_queryset = BlogPost.objects.with_likes_count_annotation()
        comments_queryset = BlogPost.objects.with_comments_count_annotation()
        tags_queryset = BlogPost.objects.with_tags_count_annotation()

        assert likes_queryset.count() == 20
        assert comments_queryset.count() == 20
        assert tags_queryset.count() == 20

        for post in likes_queryset:
            assert hasattr(post, "likes_count_annotation")

        for post in comments_queryset:
            assert hasattr(post, "comments_count_annotation")

        for post in tags_queryset:
            assert hasattr(post, "tags_count_annotation")

    def test_annotation_with_deleted_relations(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        user = UserAccountFactory()
        post.likes.add(user)
        comment = BlogCommentFactory(post=post, user=user, approved=True)
        tag = BlogTagFactory(active=True)
        post.tags.add(tag)

        likes_post = BlogPost.objects.with_likes_count_annotation().get(
            id=post.id
        )
        comments_post = BlogPost.objects.with_comments_count_annotation().get(
            id=post.id
        )
        tags_post = BlogPost.objects.with_tags_count_annotation().get(
            id=post.id
        )

        assert likes_post.likes_count_annotation == 1
        assert comments_post.comments_count_annotation == 1
        assert tags_post.tags_count_annotation == 1

        comment.delete()
        tag.delete()
        post.likes.remove(user)

        likes_post = BlogPost.objects.with_likes_count_annotation().get(
            id=post.id
        )
        comments_post = BlogPost.objects.with_comments_count_annotation().get(
            id=post.id
        )
        tags_post = BlogPost.objects.with_tags_count_annotation().get(
            id=post.id
        )

        assert likes_post.likes_count_annotation == 0
        assert comments_post.comments_count_annotation == 0
        assert tags_post.tags_count_annotation == 0

    def test_manager_get_queryset_method(self):
        manager = BlogPost.objects
        queryset = manager.get_queryset()

        assert isinstance(queryset, BlogPostQuerySet)

        assert hasattr(queryset, "with_likes_count_annotation")
        assert hasattr(queryset, "with_comments_count_annotation")
        assert hasattr(queryset, "with_tags_count_annotation")

    def test_mixed_approval_status_comments(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=True)
        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=True)
        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=False)
        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=False)
        BlogCommentFactory(post=post, user=UserAccountFactory(), approved=True)

        annotated_post = BlogPost.objects.with_comments_count_annotation().get(
            id=post.id
        )

        assert annotated_post.comments_count_annotation == 3

        assert post.comments_count == 3
        assert post.all_comments_count == 5

    def test_mixed_active_status_tags(self):
        author = BlogAuthorFactory(user=UserAccountFactory())
        post = BlogPostFactory(author=author)

        active_tag1 = BlogTagFactory(active=True)
        active_tag2 = BlogTagFactory(active=True)
        inactive_tag1 = BlogTagFactory(active=False)
        inactive_tag2 = BlogTagFactory(active=False)

        post.tags.add(active_tag1, active_tag2, inactive_tag1, inactive_tag2)

        annotated_post = BlogPost.objects.with_tags_count_annotation().get(
            id=post.id
        )

        assert annotated_post.tags_count_annotation == 2

        assert post.tags_count == 2



import uuid
from typing import Literal

from django.conf import settings
from django.contrib.postgres.indexes import BTreeIndex
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta
from djmoney.money import Money

from cart.managers.cart import CartManager
from core.models import TimeStampMixinModel, UUIDModel


class Cart(TimeStampMixinModel, UUIDModel):
    id = models.BigAutoField(primary_key=True)
    user = models.ForeignKey(
        "user.UserAccount",
        related_name="cart",
        null=True,
        blank=True,
        default=None,
        on_delete=models.CASCADE,
    )
    session_key = models.CharField(
        _("Session Key"),
        max_length=40,
        blank=True,
        help_text=_("Session key for guest users"),
        default=uuid.uuid4,
    )
    last_activity = models.DateTimeField(_("Last Activity"), auto_now=True)

    objects: CartManager = CartManager()

    class Meta(TypedModelMeta):
        verbose_name = _("Cart")
        verbose_name_plural = _("Carts")
        ordering = ["-created_at"]
        indexes = [
            *TimeStampMixinModel.Meta.indexes,
            BTreeIndex(fields=["user"], name="cart_user_ix"),
            BTreeIndex(fields=["session_key"], name="cart_session_key_ix"),
            BTreeIndex(fields=["last_activity"], name="cart_last_activity_ix"),
        ]
        constraints = [
            models.UniqueConstraint(fields=["user"], name="unique_user_cart"),
            models.UniqueConstraint(
                fields=["session_key"],
                name="unique_session_cart",
                condition=models.Q(session_key__isnull=False),
            ),
        ]

    def __str__(self):
        if self.user:
            return f"Cart for {self.user} - Items: {self.total_items} - Total: {self.total_price}"
        elif self.session_key:
            return f"Guest Cart ({self.session_key[:8]}...) - Items: {self.total_items} - Total: {self.total_price}"
        else:
            return f"Anonymous Cart {self.id} - Items: {self.total_items} - Total: {self.total_price}"

    def refresh_last_activity(self):
        self.last_activity = now()
        self.save()

    def get_items(self):
        return self.items.prefetch_related("product").all()

    @property
    def total_price(self) -> Money:
        total = sum(item.total_price.amount for item in self.get_items())
        return Money(total, settings.DEFAULT_CURRENCY)

    @property
    def total_discount_value(self) -> Money:
        total = sum(
            item.total_discount_value.amount for item in self.get_items()
        )
        return Money(total, settings.DEFAULT_CURRENCY)

    @property
    def total_vat_value(self) -> Money:
        total = sum(item.vat_value.amount for item in self.get_items())
        return Money(total, settings.DEFAULT_CURRENCY)

    @property
    def total_items(self) -> int | Literal[0]:
        return sum(item.quantity for item in self.get_items())

    @property
    def total_items_unique(self) -> int:
        return self.items.count()



from __future__ import annotations

import logging
from tempfile import NamedTemporaryFile

import requests
from allauth.account.signals import user_signed_up
from django.core.files import File
from django.dispatch import receiver

from typing import TYPE_CHECKING

from allauth.socialaccount.signals import (
    pre_social_login,
    social_account_added,
    social_account_removed,
    social_account_updated,
)
from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.utils.crypto import get_random_string

from user.models.subscription import SubscriptionTopic, UserSubscription
from user.utils.subscription import send_subscription_confirmation

if TYPE_CHECKING:  # pragma: no cover
    from allauth.socialaccount.models import SocialAccount, SocialLogin

User = get_user_model()


logger = logging.getLogger(__name__)


@receiver(user_signed_up)
def populate_profile(sociallogin=None, user=None, **kwargs):
    if not sociallogin or not user:
        logger.warning("No sociallogin or user passed to populate_profile")
        return

    provider = sociallogin.account.provider
    picture_url = None

    match provider:
        case "facebook":
            picture_url = f"http://graph.facebook.com/{sociallogin.account.uid}/picture?type=large"
        case "google":
            picture_url = sociallogin.account.extra_data.get("picture")
        case "discord":
            avatar_id = sociallogin.account.extra_data.get("id")
            avatar_hash = sociallogin.account.extra_data.get("avatar")
            if avatar_id and avatar_hash:
                picture_url = f"https://cdn.discordapp.com/avatars/{avatar_id}/{avatar_hash}.png"
            else:
                logger.warning(
                    "Missing avatar_id or avatar_hash for Discord provider"
                )
        case "github":
            picture_url = sociallogin.account.extra_data.get("avatar_url")
        case _:
            logger.warning(f"Unsupported provider: {provider}")

    if picture_url:
        try:
            response = requests.get(picture_url, timeout=10)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error(f"Failed to retrieve image from {picture_url}: {e}")
            return

        with NamedTemporaryFile(delete=True) as img_temp:
            img_temp.write(response.content)
            img_temp.flush()
            image_filename = (
                f"image_{user.first_name}_{user.last_name}_{user.pk}.jpg"
            )
            user.image.save(image_filename, File(img_temp))


@receiver(pre_social_login)
def pre_social_login_signal(
    sender, request, sociallogin: SocialLogin, **kwargs
):
    pass


@receiver(social_account_added)
def social_account_added_signal(sender, **kwargs):
    pass


@receiver(social_account_updated)
def social_account_updated_signal(
    sender, request, sociallogin: SocialLogin, **kwargs
):
    pass


@receiver(social_account_removed)
def social_account_removed_signal(
    sender, request, socialaccount: SocialAccount, **kwargs
):
    pass


@receiver(post_save, sender=User)
def create_default_subscriptions(sender, instance, created, **kwargs):
    if created:
        default_topics = SubscriptionTopic.objects.filter(
            is_active=True, is_default=True
        )

        for topic in default_topics:
            subscription = UserSubscription.objects.create(
                user=instance,
                topic=topic,
                status=(
                    UserSubscription.SubscriptionStatus.PENDING
                    if topic.requires_confirmation
                    else UserSubscription.SubscriptionStatus.ACTIVE
                ),
            )

            if topic.requires_confirmation:
                subscription.confirmation_token = get_random_string(64)
                subscription.save()
                send_subscription_confirmation(subscription, instance)



from django.core.management.base import BaseCommand

from core.caches import cache_instance


class Command(BaseCommand):
    help = "Clears cache"

    def handle(self, *args, **kwargs):
        try:
            cache_instance.clear()
            self.stdout.write(self.style.SUCCESS("Successfully cleared cache"))
        except Exception as exc:
            self.stdout.write(
                self.style.ERROR(f"Error clearing cache: {exc!s}")
            )



# Generated by Django 5.0.3 on 2024-03-29 20:53
import django.db.models.deletion
import mptt.fields
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("product", "0006_remove_product_product_search_vector_idx_and_more"),
        ("vat", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name="product",
            name="category",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="products",
                to="product.productcategory",
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="vat",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="products",
                to="vat.vat",
            ),
        ),
        migrations.AlterField(
            model_name="productreview",
            name="product",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="reviews",
                to="product.product",
            ),
        ),
        migrations.AlterField(
            model_name="productreview",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="product_reviews",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



import importlib

import factory
from django.apps import apps
from django.conf import settings
from faker import Faker

from core.factories import CustomDjangoModelFactory
from product.factories.image import ProductImageFactory
from product.factories.review import ProductReviewFactory
from product.models.product import Product
from tag.factories.tagged_item import TaggedProductFactory

fake = Faker()

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]


def get_or_create_category():
    if apps.get_model("product", "ProductCategory").objects.exists():
        return (
            apps.get_model("product", "ProductCategory")
            .objects.order_by("?")
            .first()
        )
    else:
        category_factory_module = importlib.import_module(
            "product.factories.category"
        )
        category_factory_class = category_factory_module.ProductCategoryFactory
        return category_factory_class.create()


def get_or_create_vat():
    if apps.get_model("vat", "Vat").objects.exists():
        return apps.get_model("vat", "Vat").objects.order_by("?").first()
    else:
        vat_factory_module = importlib.import_module("vat.factories")
        vat_factory_class = vat_factory_module.VatFactory
        return vat_factory_class.create()


class ProductTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    name = factory.Faker("word")
    description = factory.Faker("paragraph")
    master = factory.SubFactory("product.factories.product.ProductFactory")

    class Meta:
        model = apps.get_model("product", "ProductTranslation")
        django_get_or_create = ("language_code", "master")


class ProductFactory(CustomDjangoModelFactory):
    auto_translations = False

    unique_model_fields = [
        ("slug", lambda: fake.slug()),
    ]

    sku = factory.Faker("uuid4")
    category = factory.LazyFunction(get_or_create_category)
    price = factory.Faker(
        "pydecimal", left_digits=4, right_digits=2, positive=True
    )
    active = factory.Faker("boolean")
    stock = factory.Faker("random_int", min=5, max=100)
    discount_percent = factory.Faker(
        "pydecimal", left_digits=2, right_digits=2, positive=True, max_value=100
    )
    vat = factory.LazyFunction(get_or_create_vat)
    view_count = factory.Faker("random_int", min=0, max=1000)
    weight = factory.Faker(
        "pydecimal", left_digits=3, right_digits=2, positive=True
    )

    class Meta:
        model = Product
        django_get_or_create = ("slug",)
        skip_postgeneration_save = True

    @factory.post_generation
    def num_images(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            ProductImageFactory.create_batch(extracted, product=self)

    @factory.post_generation
    def num_reviews(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            ProductReviewFactory.create_batch(extracted, product=self)

    @factory.post_generation
    def num_tags(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            TaggedProductFactory.create_batch(extracted, content_object=self)

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        translations = extracted or [
            ProductTranslationFactory(language_code=lang, master=self)
            for lang in available_languages
        ]

        for translation in translations:
            translation.master = self
            translation.save()



import importlib
import random
from datetime import timedelta

import factory
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Count
from django.utils import timezone
from djmoney.money import Money
from faker import Faker
from phonenumber_field.phonenumber import PhoneNumber

from core.enum import FloorChoicesEnum, LocationChoicesEnum
from order.enum.document_type import OrderDocumentTypeEnum
from order.enum.status import OrderStatus, PaymentStatus
from order.factories.item import OrderItemFactory
from order.models.order import Order

User = get_user_model()

fake = Faker()


def get_or_create_user():
    if User.objects.exists():
        user = (
            User.objects.annotate(num_orders=Count("orders"))
            .order_by("num_orders")
            .first()
        )
    else:
        user_factory_module = importlib.import_module("user.factories.account")
        user_factory_class = user_factory_module.UserAccountFactory
        user = user_factory_class.create()
    return user


def get_or_create_country():
    if apps.get_model("country", "Country").objects.exists():
        return (
            apps.get_model("country", "Country").objects.order_by("?").first()
        )
    else:
        country_factory_module = importlib.import_module("country.factories")
        country_factory_class = country_factory_module.CountryFactory
        return country_factory_class.create()


def get_or_create_region():
    if apps.get_model("region", "Region").objects.exists():
        return apps.get_model("region", "Region").objects.order_by("?").first()
    else:
        region_factory_module = importlib.import_module("region.factories")
        region_factory_class = region_factory_module.RegionFactory
        return region_factory_class.create()


def get_or_create_pay_way():
    if apps.get_model("pay_way", "PayWay").objects.exists():
        return apps.get_model("pay_way", "PayWay").objects.order_by("?").first()
    else:
        pay_way_factory_module = importlib.import_module("pay_way.factories")
        pay_way_factory_class = pay_way_factory_module.PayWayFactory
        return pay_way_factory_class.create()


def generate_tracking_number():
    carriers = ["FEDEX", "UPS", "USPS", "DHL"]
    carrier = random.choice(carriers)

    if carrier == "FEDEX":
        return f"FEDEX-{fake.numerify('###########')}"
    elif carrier == "UPS":
        return f"1Z{fake.bothify('???#########')}"
    elif carrier == "USPS":
        return f"9{fake.numerify('############')}US"
    elif carrier == "DHL":
        return f"DHL-{fake.numerify('##########')}"
    return None


def generate_payment_data(status=None):
    payment_methods = ["CREDIT_CARD", "PAYPAL", "BANK_TRANSFER", "STRIPE"]
    payment_method = random.choice(payment_methods)
    payment_id: str = ""

    if payment_method == "CREDIT_CARD":
        payment_id = f"CC-{fake.numerify('###-###-###')}"
    elif payment_method == "PAYPAL":
        payment_id = f"PP-{fake.numerify('##########')}"
    elif payment_method == "BANK_TRANSFER":
        payment_id = f"BT-{fake.numerify('######')}"
    elif payment_method == "STRIPE":
        payment_id = f"pi_{fake.lexify('?' * 24)}"

    if not status:
        if random.randint(1, 10) > 2:
            status = PaymentStatus.COMPLETED
        else:
            status = random.choice(
                [
                    PaymentStatus.PENDING,
                    PaymentStatus.PROCESSING,
                    PaymentStatus.FAILED,
                ]
            )

    return {
        "payment_id": payment_id,
        "payment_method": payment_method,
        "payment_status": status,
    }


class OrderFactory(factory.django.DjangoModelFactory):
    user = factory.LazyFunction(get_or_create_user)
    country = factory.LazyFunction(get_or_create_country)
    region = factory.LazyFunction(get_or_create_region)
    pay_way = factory.LazyFunction(get_or_create_pay_way)
    floor = factory.LazyFunction(
        lambda: random.choice([s[0] for s in FloorChoicesEnum.choices])
    )
    location_type = factory.LazyFunction(
        lambda: random.choice([s[0] for s in LocationChoicesEnum.choices])
    )
    email = factory.Faker("email")
    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    street = factory.Faker("street_name")
    street_number = factory.Faker("building_number")
    city = factory.Faker("city")
    zipcode = factory.Faker("postcode")
    place = factory.LazyFunction(
        lambda: fake.secondary_address() if random.randint(1, 10) > 5 else ""
    )
    phone = factory.LazyAttribute(
        lambda _: PhoneNumber.from_string(fake.phone_number(), region="US")
    )
    mobile_phone = factory.LazyAttribute(
        lambda _: PhoneNumber.from_string(fake.phone_number(), region="US")
        if random.randint(1, 10) > 3
        else None
    )
    customer_notes = factory.LazyFunction(
        lambda: fake.paragraph(nb_sentences=2)
        if random.randint(1, 10) > 7
        else ""
    )
    status = factory.LazyFunction(
        lambda: random.choice([s[0] for s in OrderStatus.choices])
    )
    shipping_price = factory.LazyFunction(
        lambda: Money(
            fake.pydecimal(
                left_digits=2,
                right_digits=2,
                min_value=5,
                max_value=20,
                positive=True,
            ),
            settings.DEFAULT_CURRENCY,
        )
    )
    document_type = factory.LazyFunction(
        lambda: random.choice([s[0] for s in OrderDocumentTypeEnum.choices])
    )
    paid_amount = factory.LazyFunction(
        lambda: Money(
            fake.pydecimal(
                left_digits=3,
                right_digits=2,
                min_value=10,
                max_value=99,
                positive=True,
            ),
            settings.DEFAULT_CURRENCY,
        )
        if random.randint(1, 10) > 3
        else Money(0, settings.DEFAULT_CURRENCY)
    )
    payment_status = factory.LazyFunction(
        lambda: random.choice([s[0] for s in PaymentStatus.choices])
    )
    payment_id = factory.LazyFunction(
        lambda: f"payment_{fake.uuid4()}" if random.randint(1, 10) > 3 else ""
    )
    payment_method = factory.LazyFunction(
        lambda: random.choice(
            ["CREDIT_CARD", "PAYPAL", "BANK_TRANSFER", "STRIPE"]
        )
        if random.randint(1, 10) > 3
        else ""
    )
    tracking_number = factory.LazyFunction(
        lambda: generate_tracking_number() if random.randint(1, 10) > 5 else ""
    )
    shipping_carrier = factory.LazyFunction(
        lambda: random.choice(["FEDEX", "UPS", "USPS", "DHL"])
        if random.randint(1, 10) > 5
        else ""
    )
    status_updated_at = factory.LazyFunction(
        lambda: timezone.now() - timedelta(days=random.randint(0, 30))
        if random.randint(1, 10) > 3
        else None
    )

    class Meta:
        model = Order
        django_get_or_create = ("user",)
        skip_postgeneration_save = True

    @factory.post_generation
    def num_order_items(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted is not None:
            if extracted > 0:
                OrderItemFactory.create_batch(extracted, order=self)
        else:
            count = random.randint(1, 5)
            OrderItemFactory.create_batch_for_order(order=self, count=count)

    @classmethod
    def create_with_consistent_status_data(cls, status=None, **kwargs):
        if status is None:
            status = random.choice([s[0] for s in OrderStatus.choices])

        now = timezone.now()
        data = {"status": status, "status_updated_at": now, **kwargs}

        if status in [
            OrderStatus.COMPLETED,
            OrderStatus.SHIPPED,
            OrderStatus.DELIVERED,
        ]:
            payment_data = generate_payment_data(PaymentStatus.COMPLETED)
            data.update(payment_data)
            data["paid_amount"] = Money(
                fake.pydecimal(
                    left_digits=3,
                    right_digits=2,
                    min_value=10,
                    max_value=99,
                    positive=True,
                ),
                settings.DEFAULT_CURRENCY,
            )
        elif status == OrderStatus.CANCELED:
            payment_status = random.choice(
                [
                    PaymentStatus.FAILED,
                    PaymentStatus.CANCELED,
                    "",
                ]
            )
            if payment_status:
                payment_data = generate_payment_data(payment_status)
                data.update(payment_data)
        elif status == OrderStatus.REFUNDED:
            payment_data = generate_payment_data(PaymentStatus.REFUNDED)
            data.update(payment_data)
            data["paid_amount"] = Money(
                fake.pydecimal(
                    left_digits=3,
                    right_digits=2,
                    min_value=10,
                    max_value=99,
                    positive=True,
                ),
                settings.DEFAULT_CURRENCY,
            )

        if status in [OrderStatus.SHIPPED, OrderStatus.DELIVERED]:
            tracking_number = generate_tracking_number()
            carrier = (
                tracking_number.split("-")[0]
                if "-" in tracking_number
                else random.choice(["FEDEX", "UPS", "USPS", "DHL"])
            )
            data["tracking_number"] = tracking_number
            data["shipping_carrier"] = carrier

        order = cls.create(**data)
        return order

    @classmethod
    def create_pending_order(cls, **kwargs):
        return cls.create_with_consistent_status_data(
            status=OrderStatus.PENDING,
            payment_status=PaymentStatus.PENDING,
            **kwargs,
        )

    @classmethod
    def create_processing_order(cls, **kwargs):
        return cls.create_with_consistent_status_data(
            status=OrderStatus.PROCESSING,
            payment_status=PaymentStatus.COMPLETED,
            **kwargs,
        )

    @classmethod
    def create_shipped_order(cls, **kwargs):
        return cls.create_with_consistent_status_data(
            status=OrderStatus.SHIPPED, **kwargs
        )

    @classmethod
    def create_completed_order(cls, **kwargs):
        return cls.create_with_consistent_status_data(
            status=OrderStatus.COMPLETED, **kwargs
        )

    @classmethod
    def create_canceled_order(cls, **kwargs):
        return cls.create_with_consistent_status_data(
            status=OrderStatus.CANCELED, **kwargs
        )

    @classmethod
    def create_refunded_order(cls, **kwargs):
        order = cls.create_with_consistent_status_data(
            status=OrderStatus.REFUNDED,
            payment_status=PaymentStatus.REFUNDED,
            **kwargs,
        )

        if hasattr(order, "items") and order.items.exists():
            items = list(order.items.all())
            refund_all = random.choice([True, False])
            items_to_refund = (
                items
                if refund_all
                else random.sample(items, random.randint(1, len(items)))
            )

            for item in items_to_refund:
                full_refund = random.choice([True, False])
                if full_refund or item.quantity == 1:
                    refund_qty = item.quantity
                else:
                    refund_qty = random.randint(1, max(1, item.quantity - 1))

                item.refunded_quantity = refund_qty
                item.is_refunded = refund_qty == item.quantity
                item.save()

        return order



from __future__ import annotations

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.decorators import action

from rest_framework.response import Response

from core.api.serializers import ErrorResponseSerializer
from core.api.views import BaseModelViewSet

from core.utils.serializers import (
    create_schema_view_config,
    RequestSerializersConfig,
    ResponseSerializersConfig,
)
from core.utils.views import cache_methods
from notification.filters import NotificationUserFilter
from notification.models.user import NotificationUser
from notification.serializers.user import (
    NotificationCountResponseSerializer,
    NotificationSuccessResponseSerializer,
    NotificationUserActionSerializer,
    NotificationUserDetailSerializer,
    NotificationUserSerializer,
    NotificationUserWriteSerializer,
)

req_serializers: RequestSerializersConfig = {
    "create": NotificationUserWriteSerializer,
    "update": NotificationUserWriteSerializer,
    "partial_update": NotificationUserWriteSerializer,
    "mark_as_seen": NotificationUserActionSerializer,
    "mark_as_unseen": NotificationUserActionSerializer,
}

res_serializers: ResponseSerializersConfig = {
    "create": NotificationUserDetailSerializer,
    "list": NotificationUserSerializer,
    "retrieve": NotificationUserDetailSerializer,
    "update": NotificationUserDetailSerializer,
    "partial_update": NotificationUserDetailSerializer,
    "unseen_count": NotificationCountResponseSerializer,
    "mark_all_as_seen": NotificationSuccessResponseSerializer,
    "mark_all_as_unseen": NotificationSuccessResponseSerializer,
    "mark_as_seen": NotificationSuccessResponseSerializer,
    "mark_as_unseen": NotificationSuccessResponseSerializer,
}


@extend_schema_view(
    **create_schema_view_config(
        model_class=NotificationUser,
        display_config={
            "tag": "Notification Users",
        },
        request_serializers=req_serializers,
        response_serializers=res_serializers,
    ),
    unseen_count=extend_schema(
        operation_id="getNotificationUserUnseenCount",
        summary=_("Get unseen notifications count"),
        description=_(
            "Get the count of unseen notifications for the authenticated user."
        ),
        tags=["Notification Users"],
        responses={
            200: NotificationCountResponseSerializer,
            401: ErrorResponseSerializer,
        },
    ),
    mark_all_as_seen=extend_schema(
        operation_id="markAllNotificationUsersAsSeen",
        summary=_("Mark all notifications as seen"),
        description=_(
            "Mark all of the authenticated user's notifications as seen."
        ),
        tags=["Notification Users"],
        responses={
            200: NotificationSuccessResponseSerializer,
            401: ErrorResponseSerializer,
        },
    ),
    mark_all_as_unseen=extend_schema(
        operation_id="markAllNotificationUsersAsUnseen",
        summary=_("Mark all notifications as unseen"),
        description=_(
            "Mark all of the authenticated user's notifications as unseen."
        ),
        tags=["Notification Users"],
        responses={
            200: NotificationSuccessResponseSerializer,
            401: ErrorResponseSerializer,
        },
    ),
    mark_as_seen=extend_schema(
        operation_id="markNotificationUsersAsSeen",
        summary=_("Mark specific notifications as seen"),
        description=_(
            "Mark specific notifications as seen for the authenticated user."
        ),
        tags=["Notification Users"],
        request=NotificationUserActionSerializer,
        responses={
            200: NotificationSuccessResponseSerializer,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
        },
    ),
    mark_as_unseen=extend_schema(
        operation_id="markNotificationUsersAsUnseen",
        summary=_("Mark specific notifications as unseen"),
        description=_(
            "Mark specific notifications as unseen for the authenticated user."
        ),
        tags=["Notification Users"],
        request=NotificationUserActionSerializer,
        responses={
            200: NotificationSuccessResponseSerializer,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
        },
    ),
)
@cache_methods(settings.DEFAULT_CACHE_TTL, methods=["list", "retrieve"])
class NotificationUserViewSet(BaseModelViewSet):
    queryset = NotificationUser.objects.optimized_for_list()
    request_serializers = req_serializers
    response_serializers = res_serializers

    filterset_class = NotificationUserFilter
    ordering_fields = [
        "id",
        "user",
        "user__email",
        "user__first_name",
        "user__last_name",
        "notification",
        "notification__kind",
        "notification__category",
        "notification__priority",
        "notification__created_at",
        "seen",
        "seen_at",
        "created_at",
        "updated_at",
    ]
    ordering = ["-created_at", "-notification__created_at"]
    search_fields = [
        "user__first_name",
        "user__last_name",
        "user__email",
        "user__username",
        "notification__translations__title",
        "notification__translations__message",
        "notification__notification_type",
        "notification__link",
    ]

    @action(detail=False, methods=["GET"])
    def unseen_count(self, request):
        if request.user.is_anonymous:
            return Response(
                {
                    "error": _("User is not authenticated."),
                    "count": 0,
                },
                status=status.HTTP_401_UNAUTHORIZED,
            )

        count = self.queryset.filter(user=request.user, seen=False).count()
        if count == 0:
            return Response(
                {
                    "count": 0,
                },
            )

        response_data = {"count": count}

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(response_data)
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["POST"])
    def mark_all_as_seen(self, request):
        if request.user.is_anonymous:
            return Response(
                {
                    "error": _("User is not authenticated."),
                    "success": False,
                },
                status=status.HTTP_401_UNAUTHORIZED,
            )

        self.queryset.filter(user=request.user, seen=False).update(seen=True)

        response_data = {"success": True}

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(response_data)
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["POST"])
    def mark_all_as_unseen(self, request):
        if request.user.is_anonymous:
            return Response(
                {
                    "error": _("User is not authenticated."),
                    "success": False,
                },
                status=status.HTTP_401_UNAUTHORIZED,
            )

        self.queryset.filter(user=request.user, seen=True).update(seen=False)

        response_data = {"success": True}

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(response_data)
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["POST"])
    def mark_as_seen(self, request):
        if request.user.is_anonymous:
            return Response(
                {
                    "error": _("User is not authenticated."),
                    "success": False,
                },
                status=status.HTTP_401_UNAUTHORIZED,
            )

        request_serializer_class = self.get_request_serializer()
        serializer = request_serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        notification_user_ids = serializer.validated_data.get(
            "notification_user_ids"
        )

        if not notification_user_ids:
            return Response(
                {
                    "error": _("No notification user ids provided."),
                    "success": False,
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        self.queryset.filter(
            id__in=notification_user_ids, user=request.user
        ).update(seen=True)

        response_data = {"success": True}

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(response_data)
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["POST"])
    def mark_as_unseen(self, request):
        if request.user.is_anonymous:
            return Response(
                {
                    "error": _("User is not authenticated."),
                    "success": False,
                },
                status=status.HTTP_401_UNAUTHORIZED,
            )

        request_serializer_class = self.get_request_serializer()
        serializer = request_serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        notification_user_ids = serializer.validated_data.get(
            "notification_user_ids"
        )

        if not notification_user_ids:
            return Response(
                {
                    "error": _("No notification user ids provided."),
                    "success": False,
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        self.queryset.filter(
            id__in=notification_user_ids, user=request.user
        ).update(seen=False)

        response_data = {"success": True}

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(response_data)
        return Response(response_serializer.data, status=status.HTTP_200_OK)



from __future__ import annotations

from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from user.serializers.account import UserDetailsSerializer
from core.utils.testing import TestURLFixerMixin
from user.factories.account import UserAccountFactory

User = get_user_model()


class UserAccountViewSetTestCase(TestURLFixerMixin, APITestCase):
    def setUp(self):
        self.user = UserAccountFactory(
            is_superuser=True,
            is_staff=True,
        )
        self.client.force_authenticate(user=self.user)

    def get_user_account_detail_url(self, pk):
        return reverse("user-account-detail", kwargs={"pk": pk})

    def get_user_account_list_url(self):
        return reverse("user-account-list")

    def test_list(self):
        url = self.get_user_account_list_url()
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        if isinstance(response.data, dict) and "results" in response.data:
            users = User.objects.all()
            serializer = UserDetailsSerializer(
                users, many=True, context={"request": response.wsgi_request}
            )
            self.assertEqual(response.data["results"], serializer.data)
        else:
            users = User.objects.all()
            serializer = UserDetailsSerializer(
                users, many=True, context={"request": response.wsgi_request}
            )
            self.assertEqual(response.data, serializer.data)

    def test_create_valid(self):
        payload = {
            "email": "test2@test.com",
            "password": "test12345@!",
        }
        url = self.get_user_account_list_url()
        response = self.client.post(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_invalid(self):
        payload = {
            "email": "invalid_email",
            "password": "test12345@!",
        }
        url = self.get_user_account_list_url()
        response = self.client.post(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retrieve_valid(self):
        url = self.get_user_account_detail_url(self.user.id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("pk", response.data)
        self.assertEqual(response.data["pk"], self.user.id)
        self.assertIn("email", response.data)
        self.assertEqual(response.data["email"], self.user.email)
        self.assertIn("username", response.data)
        self.assertEqual(response.data["username"], self.user.username)

    def test_retrieve_invalid(self):
        invalid_user_account_id = 9999
        url = self.get_user_account_detail_url(invalid_user_account_id)
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_valid(self):
        payload = {
            "email": "test@test.com",
            "first_name": "test",
            "last_name": "test",
            "phone": "2101234567",
            "city": "test",
            "zipcode": "12345",
            "address": "test",
            "place": "test",
            "country": None,
            "region": None,
        }

        url = self.get_user_account_detail_url(self.user.pk)
        response = self.client.put(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_invalid(self):
        payload = {
            "email": "invalid_email",
            "first_name": "invalid_first_name",
            "last_name": "invalid_last_name",
            "phone": "invalid_phone",
            "city": "invalid_city",
            "zipcode": "invalid_zipcode",
            "address": "invalid_address",
            "place": "invalid_place",
            "country": "invalid_country",
            "region": "invalid_region",
        }

        url = self.get_user_account_detail_url(self.user.pk)
        response = self.client.put(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_partial_update_valid(self):
        payload = {
            "first_name": "test_partial_update",
            "last_name": "test_partial_update",
        }

        url = self.get_user_account_detail_url(self.user.pk)
        response = self.client.patch(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_partial_update_invalid(self):
        payload = {
            "email": "invalid_email",
        }

        url = self.get_user_account_detail_url(self.user.pk)
        response = self.client.patch(url, data=payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_destroy_valid(self):
        url = self.get_user_account_detail_url(self.user.pk)
        response = self.client.delete(url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(User.objects.filter(pk=self.user.pk).exists())

    def test_destroy_invalid(self):
        invalid_user_account_id = 9999
        url = self.get_user_account_detail_url(invalid_user_account_id)
        response = self.client.delete(url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



from django.urls import reverse
from rest_framework.test import APITestCase

from tag.factories.tag import TagFactory
from tag.factories.tagged_item import TaggedProductFactory
from product.factories.product import ProductFactory


class TagFilterTest(APITestCase):
    def setUp(self):
        self.tag1 = TagFactory(active=True)
        self.tag2 = TagFactory(active=False)
        self.product = ProductFactory()

        TaggedProductFactory(tag=self.tag1, content_object=self.product)

    def test_active_filter(self):
        url = reverse("tag-list")

        response = self.client.get(url, {"active": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        for tag_data in response.data["results"]:
            self.assertTrue(tag_data["active"])

        response = self.client.get(url, {"active": "false"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        for tag_data in response.data["results"]:
            self.assertFalse(tag_data["active"])

    def test_label_filters(self):
        url = reverse("tag-list")

        response = self.client.get(url, {"label__icontains": "test"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"label__icontains": "tag"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"label__istartswith": "t"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_usage_count_filters(self):
        url = reverse("tag-list")

        response = self.client.get(url, {"has_usage": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"min_usage_count": 1})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"max_usage_count": 10})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_content_type_filters(self):
        url = reverse("tag-list")

        response = self.client.get(url, {"content_type": "product"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"content_type__app_label": "product"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"object_id": self.product.id})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_timestamp_filters(self):
        url = reverse("tag-list")

        response = self.client.get(
            url, {"created_after": "2024-01-01T00:00:00Z"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(
            url, {"created_before": "2025-12-31T23:59:59Z"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_uuid_and_sort_order_filters(self):
        url = reverse("tag-list")

        response = self.client.get(url, {"uuid": str(self.tag1.uuid)})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"sort_order__gte": 1})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_camel_case_filters(self):
        url = reverse("tag-list")

        response = self.client.get(url, {"isActive": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"hasUsage": "false"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_complex_filter_combinations(self):
        url = reverse("tag-list")

        response = self.client.get(
            url,
            {"active": "true", "label__icontains": "T", "has_usage": "true"},
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(
            url, {"min_usage_count": 1, "active": "true"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_special_filters(self):
        url = reverse("tag-list")

        response = self.client.get(url, {"most_used": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"unused": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_filter_with_ordering(self):
        url = reverse("tag-list")

        response = self.client.get(
            url, {"active": "true", "ordering": "sort_order"}
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

    def test_has_label_filter(self):
        tag_without_label = TagFactory(active=True)
        tag_without_label.translations.all().delete()

        url = reverse("tag-list")

        response = self.client.get(url, {"has_label": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)

        response = self.client.get(url, {"has_label": "false"})
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data["results"], list)



from django.contrib.auth import get_user_model
from django.test import TestCase

from product.factories.favourite import ProductFavouriteFactory
from product.factories.product import ProductFactory
from user.factories.account import UserAccountFactory

User = get_user_model()


class ProductFavouriteModelTestCase(TestCase):
    def setUp(self):
        self.user = UserAccountFactory(num_addresses=0)
        self.product = ProductFactory(num_images=0, num_reviews=0)

    def test_fields(self):
        favourite = ProductFavouriteFactory(
            user=self.user, product=self.product
        )
        self.assertIsNotNone(favourite.id)
        self.assertEqual(favourite.user, self.user)
        self.assertEqual(favourite.product, self.product)

    def test_str_representation(self):
        favourite = ProductFavouriteFactory(
            user=self.user, product=self.product
        )
        self.assertEqual(
            str(favourite), f"{self.user.email} - {self.product.name}"
        )



from decimal import Decimal
from datetime import timedelta
from django.urls import reverse
from django.utils import timezone
from rest_framework.test import APITestCase

from product.enum.review import RateEnum, ReviewStatus
from product.factories.category import ProductCategoryFactory
from product.factories.product import ProductFactory
from product.models.review import ProductReview
from user.factories.account import UserAccountFactory


class ProductReviewFilterTest(APITestCase):
    def setUp(self):
        ProductReview.objects.all().delete()

        self.now = timezone.now()

        self.user1 = UserAccountFactory(
            first_name="John", last_name="Doe", email="john.doe@example.com"
        )
        self.user2 = UserAccountFactory(
            first_name="Jane", last_name="Smith", email="jane.smith@example.com"
        )
        self.prolific_user = UserAccountFactory(
            first_name="Review", last_name="Expert", email="expert@example.com"
        )

        self.admin_user = UserAccountFactory(
            first_name="Admin",
            last_name="User",
            email="admin@example.com",
            is_superuser=True,
            is_staff=True,
        )

        self.electronics_category = ProductCategoryFactory(name="Electronics")
        self.clothing_category = ProductCategoryFactory(name="Clothing")

        self.expensive_product = ProductFactory(
            category=self.electronics_category,
            price=Decimal("1000.00"),
            active=True,
        )

        self.cheap_product = ProductFactory(
            category=self.clothing_category, price=Decimal("50.00"), active=True
        )

        self.inactive_product = ProductFactory(
            category=self.electronics_category,
            price=Decimal("200.00"),
            active=False,
        )

        self.old_review = ProductReview.objects.create(
            user=self.user1,
            product=self.expensive_product,
            rate=RateEnum.FIVE,
            status=ReviewStatus.TRUE,
            is_published=True,
        )
        ProductReview.objects.filter(id=self.old_review.id).update(
            created_at=self.now - timedelta(days=30),
            updated_at=self.now - timedelta(days=25),
            published_at=self.now - timedelta(days=25),
        )
        self.old_review.refresh_from_db()
        self.old_review.set_current_language("en")
        self.old_review.comment = "Excellent product! Very satisfied with the quality and performance."
        self.old_review.save()

        self.recent_review = ProductReview.objects.create(
            user=self.user2,
            product=self.expensive_product,
            rate=RateEnum.TWO,
            status=ReviewStatus.NEW,
            is_published=False,
        )
        ProductReview.objects.filter(id=self.recent_review.id).update(
            created_at=self.now - timedelta(hours=2),
            updated_at=self.now - timedelta(hours=1),
        )
        self.recent_review.refresh_from_db()

        self.average_review = ProductReview.objects.create(
            user=self.user1,
            product=self.cheap_product,
            rate=RateEnum.THREE,
            status=ReviewStatus.TRUE,
            is_published=True,
        )
        ProductReview.objects.filter(id=self.average_review.id).update(
            created_at=self.now - timedelta(days=7),
            updated_at=self.now - timedelta(days=5),
            published_at=self.now - timedelta(days=5),
        )
        self.average_review.refresh_from_db()
        self.average_review.set_current_language("en")
        self.average_review.comment = "OK"
        self.average_review.save()

        self.fourth_review = ProductReview.objects.create(
            user=self.user2,
            product=self.cheap_product,
            rate=RateEnum.FOUR,
            status=ReviewStatus.TRUE,
            is_published=True,
        )
        ProductReview.objects.filter(id=self.fourth_review.id).update(
            created_at=self.now - timedelta(days=14),
            updated_at=self.now - timedelta(days=10),
            published_at=self.now - timedelta(days=10),
        )
        self.fourth_review.refresh_from_db()
        self.fourth_review.set_current_language("en")
        self.fourth_review.comment = (
            "Good product but had some issues with delivery."
        )
        self.fourth_review.save()

        self.prolific_review1 = ProductReview.objects.create(
            user=self.prolific_user,
            product=self.expensive_product,
            rate=RateEnum.FIVE,
            status=ReviewStatus.TRUE,
            is_published=True,
        )
        ProductReview.objects.filter(id=self.prolific_review1.id).update(
            created_at=self.now - timedelta(days=3),
            updated_at=self.now - timedelta(days=2),
            published_at=self.now - timedelta(days=2),
        )
        self.prolific_review1.refresh_from_db()

        self.prolific_review2 = ProductReview.objects.create(
            user=self.prolific_user,
            product=self.cheap_product,
            rate=RateEnum.ONE,
            status=ReviewStatus.TRUE,
            is_published=True,
        )
        ProductReview.objects.filter(id=self.prolific_review2.id).update(
            created_at=self.now - timedelta(days=1),
            updated_at=self.now - timedelta(hours=12),
            published_at=self.now - timedelta(hours=12),
        )
        self.prolific_review2.refresh_from_db()
        self.prolific_review2.set_current_language("en")
        self.prolific_review2.comment = (
            "Terrible quality, would not recommend to anyone!"
        )
        self.prolific_review2.save()

        self.client.force_authenticate(user=self.admin_user)

    def test_debug_authentication(self):
        url = reverse("product-review-list")

        self.client.force_authenticate(user=self.admin_user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        admin_count = len(response.data["results"])

        self.client.force_authenticate(user=self.user1)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        user_count = len(response.data["results"])

        self.client.force_authenticate(user=None)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        unauth_count = len(response.data["results"])

        self.client.force_authenticate(user=self.admin_user)

        self.assertGreaterEqual(admin_count, user_count)
        self.assertGreaterEqual(user_count, unauth_count)

    def test_debug_basic_api(self):
        url = reverse("product-review-list")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertIn("results", response.data)

        if response.data["results"]:
            required_fields = ["id", "rate", "status", "is_published"]
            for field in required_fields:
                self.assertIn(field, response.data["results"][0])

    def test_basic_filters(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"id": self.old_review.id})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["id"], self.old_review.id)

        response = self.client.get(url, {"user": self.user1.id})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [self.old_review.id, self.average_review.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"product": self.expensive_product.id})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [
            self.old_review.id,
            self.recent_review.id,
            self.prolific_review1.id,
        ]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

    def test_status_and_publication_filters(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"status": ReviewStatus.TRUE})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [
            self.old_review.id,
            self.average_review.id,
            self.fourth_review.id,
            self.prolific_review1.id,
            self.prolific_review2.id,
        ]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        self.client.force_authenticate(user=self.admin_user)
        response = self.client.get(url, {"status": ReviewStatus.NEW})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        if len(result_ids) > 0:
            self.assertIn(self.recent_review.id, result_ids)

        self.client.force_authenticate(user=self.user2)
        response = self.client.get(url, {"status": ReviewStatus.NEW})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        if len(result_ids) > 0:
            self.assertIn(self.recent_review.id, result_ids)

        self.client.force_authenticate(user=self.admin_user)

        response = self.client.get(url, {"is_published": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertNotIn(self.recent_review.id, result_ids)

        response = self.client.get(url, {"is_published": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        if len(result_ids) > 0:
            self.assertIn(self.recent_review.id, result_ids)

    def test_rating_filters(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"rate": RateEnum.FIVE})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [self.old_review.id, self.prolific_review1.id]
        self.assertGreaterEqual(len(result_ids), len(expected_ids))
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"rate_min": 4})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [
            self.old_review.id,
            self.fourth_review.id,
            self.prolific_review1.id,
        ]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"rate_max": 2})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [self.recent_review.id, self.prolific_review2.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"min_rate": 3, "max_rate": 4})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [self.average_review.id, self.fourth_review.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

    def test_comment_filters(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"comment": "excellent"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        if len(result_ids) > 0:
            self.assertIn(self.old_review.id, result_ids)

        response = self.client.get(url, {"has_comment": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_with_comments = [
            self.old_review.id,
            self.average_review.id,
            self.fourth_review.id,
            self.prolific_review2.id,
        ]
        found_comments = [
            rid for rid in expected_with_comments if rid in result_ids
        ]
        self.assertGreater(len(found_comments), 0)

        response = self.client.get(url, {"has_comment": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_without_comments = [
            self.recent_review.id,
            self.prolific_review1.id,
        ]
        found_no_comments = [
            rid for rid in expected_without_comments if rid in result_ids
        ]
        self.assertGreater(len(found_no_comments), 0)

    def test_related_object_filters(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"product_name": "test"})
        self.assertEqual(response.status_code, 200)

        response = self.client.get(
            url, {"product_category": self.electronics_category.id}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [
            self.old_review.id,
            self.recent_review.id,
            self.prolific_review1.id,
        ]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"product_active": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        all_count = ProductReview.objects.count()
        self.assertEqual(len(result_ids), all_count)

        response = self.client.get(url, {"user_email": "john.doe"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [self.old_review.id, self.average_review.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"user_first_name": "Jane"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [self.recent_review.id, self.fourth_review.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"user_last_name": "Expert"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        expected_ids = [self.prolific_review1.id, self.prolific_review2.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

    def test_time_based_filters(self):
        url = reverse("product-review-list")

        created_after_date = self.now - timedelta(days=14)
        response = self.client.get(
            url, {"created_after": created_after_date.isoformat()}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_recent = [
            self.recent_review.id,
            self.average_review.id,
            self.prolific_review1.id,
            self.prolific_review2.id,
        ]
        for expected_id in expected_recent:
            if (
                ProductReview.objects.get(id=expected_id).created_at
                >= created_after_date
            ):
                self.assertIn(expected_id, result_ids)

        self.assertNotIn(self.old_review.id, result_ids)

        published_after_date = self.now - timedelta(days=7)
        response = self.client.get(
            url, {"published_after": published_after_date.isoformat()}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_recently_published = [
            self.prolific_review1.id,
            self.prolific_review2.id,
        ]
        for expected_id in expected_recently_published:
            review = ProductReview.objects.get(id=expected_id)
            if (
                review.published_at
                and review.published_at >= published_after_date
            ):
                self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"recent_days": 7})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        cutoff_date = self.now - timedelta(days=7)
        for review_id in result_ids:
            review = ProductReview.objects.get(id=review_id)
            self.assertGreaterEqual(review.created_at, cutoff_date)

        response = self.client.get(url, {"published_recent_days": 3})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        cutoff_date = self.now - timedelta(days=3)
        for review_id in result_ids:
            review = ProductReview.objects.get(id=review_id)
            if review.published_at:
                self.assertGreaterEqual(review.published_at, cutoff_date)

    def test_aggregate_filters(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"user_review_count_min": 2})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        all_count = ProductReview.objects.count()
        self.assertEqual(len(result_ids), all_count)

        response = self.client.get(url, {"user_review_count_min": 3})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 0)

        response = self.client.get(url, {"product_avg_rating_min": 3})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [
            self.old_review.id,
            self.recent_review.id,
            self.prolific_review1.id,
        ]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        response = self.client.get(url, {"product_avg_rating_max": 3})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [
            self.average_review.id,
            self.fourth_review.id,
            self.prolific_review2.id,
        ]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

    def test_uuid_filter(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"uuid": str(self.old_review.uuid)})
        self.assertEqual(response.status_code, 200)

        self.assertGreaterEqual(len(response.data["results"]), 1)
        found_review = None
        for result in response.data["results"]:
            if result["id"] == self.old_review.id:
                found_review = result
                break

        self.assertIsNotNone(found_review)

    def test_camel_case_filters(self):
        url = reverse("product-review-list")

        created_after_date = self.now - timedelta(days=14)
        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "isPublished": "true",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [
            self.average_review.id,
            self.prolific_review1.id,
            self.prolific_review2.id,
        ]
        found_expected = [rid for rid in expected_ids if rid in result_ids]
        self.assertGreater(len(found_expected), 0)

        response = self.client.get(
            url,
            {
                "minRate": 3,
                "maxRate": 4,
                "hasComment": "true",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [self.average_review.id, self.fourth_review.id]
        found_expected = [rid for rid in expected_ids if rid in result_ids]
        self.assertGreater(len(found_expected), 0)

        response = self.client.get(
            url,
            {
                "userId": self.prolific_user.id,
                "currentlyPublished": "true",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [self.prolific_review1.id, self.prolific_review2.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

    def test_existing_filters_still_work(self):
        url = reverse("product-review-list")

        response = self.client.get(
            url,
            {
                "status": ReviewStatus.TRUE,
                "rate_min": 4,
                "is_published": "true",
            },
        )

        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [self.old_review.id, self.prolific_review1.id]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

        for result in response.data["results"]:
            self.assertEqual(result.get("status"), ReviewStatus.TRUE)
            self.assertGreaterEqual(result.get("rate", 0), 4)
            self.assertTrue(result.get("is_published"))

        self.assertGreaterEqual(len(result_ids), len(expected_ids))

    def test_complex_filter_combinations(self):
        url = reverse("product-review-list")

        created_after_date = self.now - timedelta(days=20)

        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "isPublished": "true",
                "minRate": 3,
                "hasComment": "true",
                "ordering": "-rate",
            },
        )
        self.assertEqual(response.status_code, 200)

        for result in response.data["results"]:
            self.assertTrue(result.get("is_published"))
            self.assertGreaterEqual(result.get("rate", 0), 3)

    def test_filter_with_ordering(self):
        url = reverse("product-review-list")

        response = self.client.get(
            url, {"status": ReviewStatus.TRUE, "ordering": "-rate"}
        )
        self.assertEqual(response.status_code, 200)

        results = response.data["results"]
        published_count = ProductReview.objects.filter(
            status=ReviewStatus.TRUE
        ).count()
        self.assertEqual(len(results), published_count)

        rates = [r.get("rate", 0) for r in results]
        self.assertEqual(rates, sorted(rates, reverse=True))

    def test_publishable_filters(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"currently_published": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]

        expected_ids = [
            self.old_review.id,
            self.average_review.id,
            self.fourth_review.id,
            self.prolific_review1.id,
            self.prolific_review2.id,
        ]
        for expected_id in expected_ids:
            self.assertIn(expected_id, result_ids)

    def test_edge_cases(self):
        url = reverse("product-review-list")

        response = self.client.get(url, {"user": 99999})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data["results"]), 0)

        response = self.client.get(url, {"rate": 0})
        self.assertEqual(response.status_code, 400)

        response = self.client.get(url, {"recent_days": 0})
        self.assertEqual(response.status_code, 200)

    def tearDown(self):
        ProductReview.objects.all().delete()



import json
import logging

from channels.generic.websocket import AsyncWebsocketConsumer
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser

logger = logging.getLogger(__name__)

User = get_user_model()


class NotificationConsumer(AsyncWebsocketConsumer):
    user: User | AnonymousUser | None = None
    group_name: str | None = None

    async def connect(self):
        try:
            logger.debug("NotificationConsumer connect called")
            self.user = self.scope["user"]
            logger.debug(
                f"User from scope: {self.user.username if not self.user.is_anonymous else 'AnonymousUser'}"
            )

            if self.user.is_anonymous:
                logger.warning("Anonymous user, closing connection")
                await self.close(code=4003)
            else:
                logger.debug(
                    f"Authenticated user: {self.user.username} (ID: {self.user.id})"
                )
                self.group_name = f"user_{self.user.id}"

                logger.debug(f"Adding user to group: {self.group_name}")
                await self.channel_layer.group_add(
                    self.group_name, self.channel_name
                )

                if self.user.is_staff:
                    logger.debug("User is staff, adding to admins group")
                    await self.channel_layer.group_add(
                        "admins", self.channel_name
                    )

                logger.debug("Accepting connection")
                await self.accept()
                logger.debug("Connection accepted")
        except KeyError as e:
            logger.error(f"KeyError in connect: {e!s}")
            await self.close(code=4000)
        except Exception as e:
            logger.exception(f"Unexpected error in connect: {e!s}")
            await self.close(code=4500)

    async def disconnect(self, close_code):
        logger.debug(
            f"NotificationConsumer disconnect called with code: {close_code}"
        )
        if not self.user:
            logger.debug("No user in disconnect")
            return

        if not self.user.is_anonymous:
            logger.debug(f"Removing user from group: {self.group_name}")
            await self.channel_layer.group_discard(
                self.group_name, self.channel_name
            )

            if self.user.is_staff:
                logger.debug("User is staff, removing from admins group")
                await self.channel_layer.group_discard(
                    "admins", self.channel_name
                )

    async def send_notification(self, event):
        logger.debug(f"Sending notification: {event}")
        await self.send(text_data=json.dumps(event))

    async def receive(self, text_data=None, bytes_data=None):
        logger.debug(f"Received message: {text_data}")
        try:
            data = json.loads(text_data)
            await self.send(
                text_data=json.dumps(
                    {
                        "type": "echo.message",
                        "message": data,
                        "from": self.user.username
                        if self.user and not self.user.is_anonymous
                        else "anonymous",
                    }
                )
            )
        except json.JSONDecodeError:
            logger.error(f"Invalid JSON received: {text_data}")
            await self.send(
                text_data=json.dumps(
                    {"type": "error", "message": "Invalid JSON format"}
                )
            )



import importlib

import factory
from django.apps import apps
from django.conf import settings
from faker import Faker

from core.factories import (
    CustomDjangoModelFactory,
    custom_seeding,
    SeedingResult,
)
from region.models import Region

fake = Faker()

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]

COUNTRY_REGIONS_MAPPING = getattr(settings, "COUNTRY_REGIONS_MAPPING", {})


def get_or_create_country():
    if apps.get_model("country", "Country").objects.exists():
        return (
            apps.get_model("country", "Country").objects.order_by("?").first()
        )
    else:
        country_factory_module = importlib.import_module("country.factories")
        country_factory_class = country_factory_module.CountryFactory
        return country_factory_class.create()


class RegionTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    name = factory.LazyAttribute(
        lambda obj: f"Region Name ({obj.language_code})"
    )
    master = factory.SubFactory("region.factories.RegionFactory")

    class Meta:
        model = apps.get_model("region", "RegionTranslation")
        django_get_or_create = ("language_code", "master")


@custom_seeding(
    description="Creates regions based on COUNTRY_REGIONS_MAPPING for existing countries",
    settings_key="COUNTRY_REGIONS_MAPPING",
)
class RegionFactory(CustomDjangoModelFactory):
    auto_translations = False

    unique_model_fields = [
        ("alpha", lambda: fake.bothify(text="???-######")),
    ]

    country = factory.LazyFunction(get_or_create_country)

    class Meta:
        model = Region
        django_get_or_create = ("alpha",)
        skip_postgeneration_save = True

    @classmethod
    def custom_seed(cls, **kwargs) -> SeedingResult:
        result = SeedingResult(created_count=0)
        verbose = kwargs.get("verbose", True)

        from country.models import Country

        for country_code, regions_data in COUNTRY_REGIONS_MAPPING.items():
            try:
                country = Country.objects.get(alpha_2=country_code)
            except Country.DoesNotExist:
                if verbose:
                    print(
                        f"Country {country_code} not found, skipping regions..."
                    )
                result.skipped_count += 1
                continue

            for region_data in regions_data:
                existing_region = Region.objects.filter(
                    alpha=region_data["alpha"]
                ).first()

                if existing_region:
                    if verbose:
                        print(
                            f"Region {region_data['alpha']} already exists, skipping..."
                        )
                    result.skipped_count += 1
                    continue

                try:
                    region = Region.objects.create(
                        alpha=region_data["alpha"],
                        country=country,
                    )

                    for trans_lang in available_languages:
                        region_name = region_data["names"].get(
                            trans_lang,
                            region_data["names"].get(
                                "en", region_data["alpha"]
                            ),
                        )
                        translation_model = apps.get_model(
                            "region", "RegionTranslation"
                        )
                        translation_model.objects.get_or_create(
                            language_code=trans_lang,
                            master=region,
                            defaults={"name": region_name},
                        )

                    result.created_count += 1
                    if verbose:
                        print(
                            f"Created region: {region_data['alpha']} in {country_code}"
                        )

                except Exception as e:
                    error_msg = f"Failed to create region {region_data['alpha']}: {str(e)}"
                    result.errors.append(error_msg)
                    if verbose:
                        print(f"Error: {error_msg}")

        return result

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        if not hasattr(self, "_translations_created"):
            translations = extracted or [
                RegionTranslationFactory(language_code=lang, master=self)
                for lang in available_languages
            ]

            for translation in translations:
                translation.save()



from parler.managers import TranslatableManager, TranslatableQuerySet


class PayWayQuerySet(TranslatableQuerySet):
    def active(self):
        return self.filter(active=True)

    def inactive(self):
        return self.filter(active=False)

    def online_payments(self):
        return self.filter(is_online_payment=True)

    def offline_payments(self):
        return self.filter(is_online_payment=False)


class PayWayManager(TranslatableManager):
    def get_queryset(self) -> PayWayQuerySet:
        return PayWayQuerySet(self.model, using=self._db)

    def active(self):
        return self.get_queryset().active()

    def inactive(self):
        return self.get_queryset().inactive()

    def online_payments(self):
        return self.get_queryset().online_payments()

    def offline_payments(self):
        return self.get_queryset().offline_payments()



# Generated by Django 5.2.2 on 2025-06-08 22:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tag', '0004_remove_taggeditem_tag_taggedi_content_cd6f88_idx_and_more'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='tag',
            managers=[
            ],
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-10 14:39
import uuid

import django.db.models.deletion
import phonenumber_field.modelfields
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("country", "0001_initial"),
        ("region", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserAccount",
            fields=[
                (
                    "password",
                    models.CharField(max_length=128, verbose_name="password"),
                ),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created At"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="Updated At"),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "email",
                    models.EmailField(
                        max_length=254,
                        unique=True,
                        verbose_name="Email Address",
                    ),
                ),
                (
                    "first_name",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="First Name",
                    ),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Last Name",
                    ),
                ),
                (
                    "phone",
                    phonenumber_field.modelfields.PhoneNumberField(
                        blank=True,
                        default=None,
                        max_length=128,
                        null=True,
                        region=None,
                        verbose_name="Phone Number",
                    ),
                ),
                (
                    "city",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="City",
                    ),
                ),
                (
                    "zipcode",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Zip Code",
                    ),
                ),
                (
                    "address",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Address",
                    ),
                ),
                (
                    "place",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Place",
                    ),
                ),
                (
                    "image",
                    models.ImageField(
                        blank=True,
                        null=True,
                        upload_to="uploads/users/",
                        verbose_name="Image",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(default=True, verbose_name="Active"),
                ),
                (
                    "is_staff",
                    models.BooleanField(default=False, verbose_name="Staff"),
                ),
                (
                    "birth_date",
                    models.DateField(blank=True, null=True, verbose_name="Birth Date"),
                ),
                (
                    "twitter",
                    models.URLField(
                        blank=True, null=True, verbose_name="Twitter Profile"
                    ),
                ),
                (
                    "linkedin",
                    models.URLField(
                        blank=True, null=True, verbose_name="LinkedIn Profile"
                    ),
                ),
                (
                    "facebook",
                    models.URLField(
                        blank=True, null=True, verbose_name="Facebook Profile"
                    ),
                ),
                (
                    "instagram",
                    models.URLField(
                        blank=True, null=True, verbose_name="Instagram Profile"
                    ),
                ),
                (
                    "website",
                    models.URLField(blank=True, null=True, verbose_name="Website"),
                ),
                (
                    "youtube",
                    models.URLField(
                        blank=True, null=True, verbose_name="Youtube Profile"
                    ),
                ),
                (
                    "github",
                    models.URLField(
                        blank=True, null=True, verbose_name="Github Profile"
                    ),
                ),
                (
                    "bio",
                    models.TextField(blank=True, null=True, verbose_name="Bio"),
                ),
                (
                    "country",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="user_account_country",
                        to="country.country",
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all"
                        " permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "region",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="user_account_region",
                        to="region.region",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "verbose_name": "User Account",
                "verbose_name_plural": "User Accounts",
                "ordering": ["-created_at"],
            },
        ),
        migrations.CreateModel(
            name="UserAddress",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created At"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="Updated At"),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "title",
                    models.CharField(max_length=255, verbose_name="Title"),
                ),
                (
                    "first_name",
                    models.CharField(max_length=255, verbose_name="First Name"),
                ),
                (
                    "last_name",
                    models.CharField(max_length=255, verbose_name="Last Name"),
                ),
                (
                    "street",
                    models.CharField(max_length=255, verbose_name="Street"),
                ),
                (
                    "street_number",
                    models.CharField(max_length=255, verbose_name="Street Number"),
                ),
                (
                    "city",
                    models.CharField(max_length=255, verbose_name="City"),
                ),
                (
                    "zipcode",
                    models.CharField(max_length=255, verbose_name="Zip Code"),
                ),
                (
                    "floor",
                    models.PositiveSmallIntegerField(
                        blank=True,
                        choices=[
                            (0, "Basement"),
                            (1, "Ground Floor"),
                            (2, "First Floor"),
                            (3, "Second Floor"),
                            (4, "Third Floor"),
                            (5, "Fourth Floor"),
                            (6, "Fifth Floor"),
                            (7, "Sixth Floor Plus"),
                        ],
                        default=None,
                        null=True,
                        verbose_name="Floor",
                    ),
                ),
                (
                    "location_type",
                    models.PositiveSmallIntegerField(
                        blank=True,
                        choices=[(0, "Home"), (1, "Office"), (2, "Other")],
                        default=None,
                        null=True,
                        verbose_name="Location Type",
                    ),
                ),
                (
                    "phone",
                    phonenumber_field.modelfields.PhoneNumberField(
                        blank=True,
                        default=None,
                        max_length=128,
                        null=True,
                        region=None,
                        verbose_name="Phone Number",
                    ),
                ),
                (
                    "mobile_phone",
                    phonenumber_field.modelfields.PhoneNumberField(
                        max_length=128,
                        region=None,
                        verbose_name="Mobile Phone Number",
                    ),
                ),
                (
                    "notes",
                    models.CharField(
                        blank=True,
                        default=None,
                        max_length=255,
                        null=True,
                        verbose_name="Notes",
                    ),
                ),
                (
                    "is_main",
                    models.BooleanField(default=False, verbose_name="Is Main"),
                ),
                (
                    "country",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="address_country",
                        to="country.country",
                    ),
                ),
                (
                    "region",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="address_region",
                        to="region.region",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="user_address",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "User Address",
                "verbose_name_plural": "User Addresses",
                "ordering": ["-created_at"],
            },
        ),
        migrations.AddConstraint(
            model_name="useraddress",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_main", True)),
                fields=("user", "is_main"),
                name="unique_main_address",
            ),
        ),
    ]



from __future__ import annotations

import logging
from rest_framework.response import Response
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound, PermissionDenied
from rest_framework.exceptions import ValidationError as DRFValidationError
from rest_framework import status
from rest_framework.permissions import IsAuthenticated

from core.api.serializers import ErrorResponseSerializer
from core.api.views import BaseModelViewSet

from core.utils.serializers import (
    create_schema_view_config,
    RequestSerializersConfig,
    ResponseSerializersConfig,
)
from core.utils.views import cache_methods
from order.enum.status import OrderStatus
from order.filters import OrderItemFilter
from order.models.item import OrderItem
from order.serializers.item import (
    OrderItemDetailSerializer,
    OrderItemRefundSerializer,
    OrderItemSerializer,
    OrderItemWriteSerializer,
    OrderItemRefundResponseSerializer,
)

req_serializers: RequestSerializersConfig = {
    "create": OrderItemWriteSerializer,
    "update": OrderItemWriteSerializer,
    "partial_update": OrderItemWriteSerializer,
    "refund": OrderItemRefundSerializer,
}

res_serializers: ResponseSerializersConfig = {
    "create": OrderItemDetailSerializer,
    "list": OrderItemSerializer,
    "retrieve": OrderItemDetailSerializer,
    "update": OrderItemDetailSerializer,
    "partial_update": OrderItemDetailSerializer,
    "refund": OrderItemRefundResponseSerializer,
}


@extend_schema_view(
    **create_schema_view_config(
        model_class=OrderItem,
        display_config={
            "tag": "Order Items",
        },
        request_serializers=req_serializers,
        response_serializers=res_serializers,
    ),
    refund=extend_schema(
        operation_id="refundOrderItem",
        summary=_("Process a refund for an order item"),
        description=_("Process a refund for an order item."),
        tags=["Order Items"],
        request=OrderItemRefundSerializer,
        responses={
            200: OrderItemRefundResponseSerializer,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
            403: ErrorResponseSerializer,
            404: ErrorResponseSerializer,
        },
    ),
)
@cache_methods(settings.DEFAULT_CACHE_TTL, methods=["list", "retrieve"])
class OrderItemViewSet(BaseModelViewSet):
    queryset = OrderItem.objects.none()
    request_serializers = req_serializers
    response_serializers = res_serializers
    permission_classes = [IsAuthenticated]
    filterset_class = OrderItemFilter
    ordering_fields = [
        "id",
        "created_at",
        "updated_at",
        "quantity",
        "price",
        "sort_order",
    ]
    ordering = ["-created_at"]
    search_fields = [
        "product__translations__name",
        "notes",
        "order__first_name",
        "order__last_name",
        "order__email",
    ]

    def get_queryset(self):
        user = self.request.user

        if user.is_staff or user.is_superuser:
            return (
                OrderItem.objects.all()
                .select_related("order", "product")
                .prefetch_related("product__translations")
            )

        return (
            OrderItem.objects.filter(order__user=user)
            .select_related("order", "product")
            .prefetch_related("product__translations")
        )

    def get_object(self):
        try:
            obj = super().get_object()
            self.check_order_permission(obj.order)
            return obj
        except OrderItem.DoesNotExist as e:
            raise NotFound(_("Order item not found.")) from e

    def check_order_permission(self, order):
        user = self.request.user

        if user.is_staff or user.is_superuser:
            return

        if not order.user or order.user.id != user.id:
            raise PermissionDenied(
                _("You do not have permission to access this order.")
            )

    def perform_create(self, serializer):
        product = serializer.validated_data["product"]
        if not serializer.validated_data.get("price"):
            serializer.save(price=product.price)
        else:
            serializer.save()

    @action(detail=True, methods=["POST"])
    def refund(self, request, pk=None):
        order_item = self.get_object()

        if order_item.order.status not in [
            OrderStatus.DELIVERED,
            OrderStatus.COMPLETED,
            OrderStatus.RETURNED,
        ]:
            raise DRFValidationError(
                {"detail": _("This order is not in a refundable state.")}
            )

        if order_item.is_refunded:
            raise DRFValidationError(
                {"detail": _("This item has already been fully refunded.")}
            )

        request_serializer_class = self.get_request_serializer()
        serializer = request_serializer_class(
            data=request.data,
            context={"item": order_item, **self.get_serializer_context()},
        )
        serializer.is_valid(raise_exception=True)

        quantity = serializer.validated_data.get("quantity")
        reason = serializer.validated_data.get("reason", "")

        try:
            if reason:
                notes = order_item.notes or ""
                order_item.notes = f"{notes}\nRefund reason: {reason}".strip()
                order_item.save(update_fields=["notes"])

            refunded_amount = order_item.refund(quantity)

            response_data = {
                "detail": _("Refund processed successfully."),
                "refunded_amount": refunded_amount,
                "item": order_item,
            }

            response_serializer_class = self.get_response_serializer()
            response_serializer = response_serializer_class(response_data)
            return Response(response_serializer.data, status=status.HTTP_200_OK)

        except ValidationError as e:
            logger = logging.getLogger(__name__)

            logger.error(
                f"Validation error during refund: {e!s}",
                extra={
                    "order_item_id": order_item.id,
                    "user_id": request.user.id,
                },
            )
            raise DRFValidationError(
                {
                    "detail": _(
                        "A validation error occurred while processing the refund. Please check your input and try again."
                    )
                }
            ) from e
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.error(
                f"Error processing refund: {e!s}",
                extra={
                    "order_item_id": order_item.id,
                    "user_id": request.user.id,
                },
            )
            raise DRFValidationError(
                {
                    "detail": _(
                        "An error occurred while processing the refund. Please try again later."
                    )
                }
            ) from e



from __future__ import annotations

import os

from django.contrib.postgres.indexes import BTreeIndex
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta
from parler.models import TranslatableModel, TranslatedFields

from core.fields.image import ImageAndSvgField
from core.models import SortableModel, TimeStampMixinModel, UUIDModel
from product.managers.image import EnhancedImageManager


class ProductImage(
    TranslatableModel, TimeStampMixinModel, SortableModel, UUIDModel
):
    id = models.BigAutoField(primary_key=True)
    product = models.ForeignKey(
        "product.Product",
        related_name="images",
        on_delete=models.CASCADE,
    )
    image = ImageAndSvgField(_("Image"), upload_to="uploads/products/")
    is_main = models.BooleanField(_("Is Main"), default=False)
    translations = TranslatedFields(
        title=models.CharField(_("Title"), max_length=50, blank=True, null=True)
    )

    objects: EnhancedImageManager = EnhancedImageManager()

    class Meta(TypedModelMeta):
        verbose_name = _("Product Image")
        verbose_name_plural = _("Product Images")
        ordering = ["sort_order"]
        indexes = [
            *TimeStampMixinModel.Meta.indexes,
            *SortableModel.Meta.indexes,
            BTreeIndex(
                fields=["product", "is_main"],
                name="prod_img_product_is_main_ix",
            ),
        ]

    def __str__(self):
        product_name = self.product.safe_translation_getter(
            "name", any_language=True
        )
        main_status = "Main" if self.is_main else "Secondary"
        return f"{product_name} Image ({main_status})"

    def get_ordering_queryset(self):
        return self.product.images.all()

    def clean(self):
        if self.is_main:
            ProductImage.objects.filter(
                product=self.product, is_main=True
            ).update(is_main=False)
        super().clean()

    @property
    def main_image_path(self) -> str:
        if self.image and hasattr(self.image, "name"):
            return f"media/uploads/products/{os.path.basename(self.image.name)}"
        return ""



from django.contrib.postgres.indexes import BTreeIndex, GinIndex
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta
from phonenumber_field.modelfields import PhoneNumberField

from core.enum import FloorChoicesEnum, LocationChoicesEnum
from core.models import TimeStampMixinModel, UUIDModel
from user.managers.address import UserAddressManager


class UserAddress(TimeStampMixinModel, UUIDModel):
    id = models.BigAutoField(primary_key=True)
    user = models.ForeignKey(
        "user.UserAccount",
        related_name="addresses",
        on_delete=models.CASCADE,
    )
    title = models.CharField(_("Title"), max_length=255)
    first_name = models.CharField(_("First Name"), max_length=255)
    last_name = models.CharField(_("Last Name"), max_length=255)
    street = models.CharField(_("Street"), max_length=255)
    street_number = models.CharField(_("Street Number"), max_length=255)
    city = models.CharField(_("City"), max_length=255)
    zipcode = models.CharField(_("Zip Code"), max_length=255)
    country = models.ForeignKey(
        "country.Country",
        related_name="user_addresses",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
    )
    region = models.ForeignKey(
        "region.Region",
        related_name="user_addresses",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
    )
    floor = models.CharField(
        _("Floor"),
        max_length=50,
        choices=FloorChoicesEnum,
        blank=True,
        default="",
    )
    location_type = models.CharField(
        _("Location Type"),
        max_length=100,
        choices=LocationChoicesEnum,
        blank=True,
        default="",
    )
    phone = PhoneNumberField(
        _("Phone Number"), null=True, blank=True, default=None
    )
    mobile_phone = PhoneNumberField(_("Mobile Phone Number"))
    notes = models.CharField(_("Notes"), max_length=255, blank=True, default="")
    is_main = models.BooleanField(_("Is Main"), default=False)

    objects: UserAddressManager = UserAddressManager()

    class Meta(TypedModelMeta):
        verbose_name = _("User Address")
        verbose_name_plural = _("User Addresses")
        ordering = ["-is_main"]
        constraints = [
            models.UniqueConstraint(
                fields=["user", "is_main"],
                condition=models.Q(is_main=True),
                name="unique_main_address",
            )
        ]
        indexes = [
            *TimeStampMixinModel.Meta.indexes,
            BTreeIndex(fields=["user"], name="user_address_user_ix"),
            BTreeIndex(fields=["is_main"], name="user_address_is_main_ix"),
            BTreeIndex(fields=["country"], name="user_address_country_ix"),
            BTreeIndex(fields=["region"], name="user_address_region_ix"),
            GinIndex(
                name="user_address_search_gin",
                fields=["title", "first_name", "last_name", "city"],
                opclasses=["gin_trgm_ops"] * 4,
            ),
        ]

    def __str__(self):
        return f"{self.title} - {self.first_name} {self.last_name}, {self.city}"

    def save(self, *args, **kwargs):
        if self.is_main:
            UserAddress.objects.filter(user=self.user, is_main=True).exclude(
                pk=self.id
            ).update(is_main=False)
        super().save(*args, **kwargs)

    def clean(self):
        if self.is_main:
            main_count = (
                UserAddress.objects.filter(user=self.user, is_main=True)
                .exclude(pk=self.pk)
                .count()
            )
            if main_count > 0:
                raise ValidationError(
                    _("There can only be one main address per user.")
                )

    @classmethod
    def get_user_addresses(cls, user):
        return cls.objects.filter(user=user)

    @classmethod
    def get_main_address(cls, user):
        return cls.objects.filter(user=user, is_main=True).first()

    @classmethod
    def get_user_address_count(cls, user):
        return cls.objects.filter(user=user).count()



import pytest
from decimal import Decimal
from unittest.mock import Mock, patch

from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.test import RequestFactory

from product.admin import (
    ProductCategoryAdmin,
    ProductAdmin,
    ProductReviewAdmin,
    ProductFavouriteAdmin,
    ProductCategoryImageAdmin,
    ProductImageAdmin,
    StockStatusFilter,
    PriceRangeFilter,
    DiscountStatusFilter,
    PopularityFilter,
    LikesCountFilter,
    ReviewAverageFilter,
    ProductImageInline,
    ProductCategoryImageInline,
)
from product.factories import (
    ProductFactory,
    ProductCategoryFactory,
    ProductReviewFactory,
    ProductFavouriteFactory,
    ProductImageFactory,
    ProductCategoryImageFactory,
)
from product.models import (
    Product,
    ProductCategory,
    ProductReview,
    ProductFavourite,
    ProductImage,
    ProductCategoryImage,
)
from tag.admin import TaggedItemInline
from user.factories import UserAccountFactory

User = get_user_model()


@pytest.fixture
def admin_request():
    factory = RequestFactory()
    request = factory.get("/admin/")
    request.user = Mock()
    request.user.is_authenticated = True
    request.user.is_staff = True
    request.user.is_superuser = True
    return request


@pytest.fixture
def product_admin():
    return ProductAdmin(Product, AdminSite())


@pytest.fixture
def category_admin():
    return ProductCategoryAdmin(ProductCategory, AdminSite())


@pytest.fixture
def review_admin():
    return ProductReviewAdmin(ProductReview, AdminSite())


@pytest.fixture
def favourite_admin():
    return ProductFavouriteAdmin(ProductFavourite, AdminSite())


@pytest.fixture
def product_image_admin():
    return ProductImageAdmin(ProductImage, AdminSite())


@pytest.fixture
def category_image_admin():
    return ProductCategoryImageAdmin(ProductCategoryImage, AdminSite())


@pytest.mark.django_db
class TestStockStatusFilter:
    def test_filter_lookups(self, admin_request):
        filter_instance = StockStatusFilter(
            admin_request, {}, Product, ProductAdmin(Product, AdminSite())
        )

        lookups = filter_instance.lookups(
            admin_request, ProductAdmin(Product, AdminSite())
        )

        assert len(lookups) == 5
        lookup_values = [lookup[0] for lookup in lookups]
        assert "in_stock" in lookup_values
        assert "low_stock" in lookup_values
        assert "out_of_stock" in lookup_values
        assert "high_stock" in lookup_values
        assert "critical_stock" in lookup_values

    def test_filter_queryset_in_stock(self, admin_request):
        ProductFactory(stock=10)
        ProductFactory(stock=0)
        ProductFactory(stock=25)

        queryset = Product.objects.all()

        filtered_queryset = queryset.filter(stock__gt=0)

        assert filtered_queryset.count() == 2
        for product in filtered_queryset:
            assert product.stock > 0

    def test_filter_queryset_out_of_stock(self, admin_request):
        ProductFactory(stock=10)
        ProductFactory(stock=0)
        ProductFactory(stock=0)

        queryset = Product.objects.all()
        filtered_queryset = queryset.filter(stock=0)

        assert filtered_queryset.count() == 2
        for product in filtered_queryset:
            assert product.stock == 0


@pytest.mark.django_db
class TestPriceRangeFilter:
    def test_filter_lookups(self, admin_request):
        filter_instance = PriceRangeFilter(
            admin_request, {}, Product, ProductAdmin(Product, AdminSite())
        )

        lookups = filter_instance.lookups(
            admin_request, ProductAdmin(Product, AdminSite())
        )

        assert len(lookups) == 5
        lookup_values = [lookup[0] for lookup in lookups]
        assert "budget" in lookup_values
        assert "luxury" in lookup_values

    def test_filter_queryset_budget(self, admin_request):
        ProductFactory(price=Decimal("15.00"))
        ProductFactory(price=Decimal("75.00"))
        ProductFactory(price=Decimal("600.00"))

        queryset = Product.objects.all()
        filtered_queryset = queryset.filter(price__lte=20)

        assert filtered_queryset.count() == 1
        assert float(filtered_queryset.first().price.amount) <= 20


@pytest.mark.django_db
class TestDiscountStatusFilter:
    def test_filter_lookups(self, admin_request):
        filter_instance = DiscountStatusFilter(
            admin_request, {}, Product, ProductAdmin(Product, AdminSite())
        )

        lookups = filter_instance.lookups(
            admin_request, ProductAdmin(Product, AdminSite())
        )

        assert len(lookups) == 4
        lookup_values = [lookup[0] for lookup in lookups]
        assert "on_sale" in lookup_values
        assert "no_discount" in lookup_values

    def test_filter_queryset_on_sale(self, admin_request):
        ProductFactory(discount_percent=15)
        ProductFactory(discount_percent=0)
        ProductFactory(discount_percent=25)

        queryset = Product.objects.all()
        filtered_queryset = queryset.filter(discount_percent__gt=0)

        assert filtered_queryset.count() == 2
        for product in filtered_queryset:
            assert product.discount_percent > 0


@pytest.mark.django_db
class TestPopularityFilter:
    def test_filter_lookups(self, admin_request):
        filter_instance = PopularityFilter(
            admin_request, {}, Product, ProductAdmin(Product, AdminSite())
        )

        lookups = filter_instance.lookups(
            admin_request, ProductAdmin(Product, AdminSite())
        )

        assert len(lookups) == 4
        lookup_values = [lookup[0] for lookup in lookups]
        assert "trending" in lookup_values
        assert "new_arrivals" in lookup_values

    def test_filter_queryset_trending(self, admin_request):
        ProductFactory(view_count=150)
        ProductFactory(view_count=50)
        ProductFactory(view_count=200)

        queryset = Product.objects.all()
        filtered_queryset = queryset.filter(view_count__gt=100)

        assert filtered_queryset.count() == 2
        for product in filtered_queryset:
            assert product.view_count > 100


@pytest.mark.django_db
class TestLikesCountFilter:
    def test_filter_expected_parameters(self, admin_request):
        filter_instance = LikesCountFilter(
            admin_request, {}, Product, ProductAdmin(Product, AdminSite())
        )

        expected = filter_instance.expected_parameters()

        assert "likes_count_from" in expected
        assert "likes_count_to" in expected


@pytest.mark.django_db
class TestReviewAverageFilter:
    def test_filter_expected_parameters(self, admin_request):
        filter_instance = ReviewAverageFilter(
            admin_request, {}, Product, ProductAdmin(Product, AdminSite())
        )

        expected = filter_instance.expected_parameters()

        assert "review_average_from" in expected
        assert "review_average_to" in expected


@pytest.mark.django_db
class TestProductImageInline:
    def test_inline_configuration(self, admin_request):
        inline = ProductImageInline(Product, AdminSite())

        assert inline.model == ProductImage
        assert inline.extra == 0
        assert "image_preview" in inline.fields
        assert "image_preview" in inline.readonly_fields

    def test_image_preview_with_image(self, admin_request):
        product = ProductFactory()
        product_image = ProductImageFactory(product=product)
        inline = ProductImageInline(Product, AdminSite())

        result = inline.image_preview(product_image)

        assert len(result) > 0


@pytest.mark.django_db
class TestTaggedItemInline:
    def test_inline_configuration(self, admin_request):
        inline = TaggedItemInline(Product, AdminSite())

        assert inline.extra == 0
        assert "tag" in inline.fields
        assert inline.ct_field == "content_type"
        assert inline.ct_fk_field == "object_id"


@pytest.mark.django_db
class TestProductCategoryImageInline:
    def test_inline_configuration(self, admin_request):
        inline = ProductCategoryImageInline(ProductCategory, AdminSite())

        assert inline.model == ProductCategoryImage
        assert inline.extra == 0
        assert "image_preview" in inline.fields
        assert "image_preview" in inline.readonly_fields


@pytest.mark.django_db
class TestCategoryAdmin:
    def test_category_info_display(self, category_admin):
        category = ProductCategoryFactory()

        result = category_admin.category_info(category)

        assert (
            category.safe_translation_getter("name", any_language=True)
            in result
        )
        assert "Level:" in result
        assert "font-medium" in result

    def test_category_stats_display(self, category_admin):
        category = ProductCategoryFactory()

        result = category_admin.category_stats(category)

        assert "Direct:" in result
        assert "Total:" in result
        assert "Subcategories:" in result

    def test_category_status_active(self, category_admin):
        category = ProductCategoryFactory(active=True)

        result = category_admin.category_status(category)

        assert "✅" in result
        assert "Active" in result

    def test_category_status_inactive(self, category_admin):
        category = ProductCategoryFactory(active=False)

        result = category_admin.category_status(category)

        assert "❌" in result
        assert "Inactive" in result

    def test_image_preview_with_image(self, category_admin):
        category = ProductCategoryFactory()

        result = category_admin.image_preview(category)

        assert len(result) >= 0

    def test_created_display(self, category_admin):
        category = ProductCategoryFactory()

        result = category_admin.created_display(category)

        assert len(result) > 10
        assert "2025" in result or "2024" in result

    def test_category_analytics(self, category_admin):
        category = ProductCategoryFactory()

        result = category_admin.category_analytics(category)

        assert "Level:" in result
        assert "Direct Products:" in result
        assert "Total Products:" in result


@pytest.mark.django_db
class TestProductAdmin:
    def test_product_info_display(self, product_admin):
        product = ProductFactory()

        result = product_admin.product_info(product)

        assert (
            product.safe_translation_getter("name", any_language=True) in result
        )
        assert product.sku[:8] in result

    def test_category_display(self, product_admin):
        category = ProductCategoryFactory()
        product = ProductFactory(category=category)

        result = product_admin.category_display(product)

        assert (
            category.safe_translation_getter("name", any_language=True)
            in result
        )

    def test_pricing_info_display(self, product_admin):
        product = ProductFactory(price=Decimal("99.99"), discount_percent=20)

        result = product_admin.pricing_info(product)

        assert "99,99" in result
        assert "€" in result
        assert "20%" in result

    def test_stock_info_display(self, product_admin):
        product = ProductFactory(stock=25)

        result = product_admin.stock_info(product)

        assert "25" in result
        assert "In Stock" in result

    def test_performance_metrics_display(self, product_admin):
        product = ProductFactory(view_count=150)

        result = product_admin.performance_metrics(product)

        assert "150" in result
        assert "👁️" in result

    def test_status_badges_active(self, product_admin):
        product = ProductFactory(active=True)

        result = product_admin.status_badges(product)

        assert "✅" in result
        assert "Active" in result

    def test_status_badges_inactive(self, product_admin):
        product = ProductFactory(active=False)

        result = product_admin.status_badges(product)

        assert "❌" in result
        assert "Inactive" in result

    def test_created_display(self, product_admin):
        product = ProductFactory()

        result = product_admin.created_display(product)

        assert len(result) > 10
        assert "2025" in result or "2024" in result

    def test_pricing_summary(self, product_admin):
        product = ProductFactory(price=Decimal("100.00"), discount_percent=10)

        result = product_admin.pricing_summary(product)

        assert "Base Price" in result
        assert "100,00" in result

    def test_performance_summary(self, product_admin):
        product = ProductFactory(view_count=100)

        result = product_admin.performance_summary(product)

        assert "Total Views" in result
        assert "100" in result

    def test_product_analytics(self, product_admin):
        product = ProductFactory()

        result = product_admin.product_analytics(product)

        assert "Product Age" in result
        assert "Engagement" in result

    @patch.object(ProductAdmin, "message_user")
    def test_make_active_action(
        self, mock_message, product_admin, admin_request
    ):
        products = [ProductFactory(active=False) for _ in range(3)]
        queryset = Product.objects.filter(id__in=[p.id for p in products])

        product_admin.make_active(admin_request, queryset)

        for product in products:
            product.refresh_from_db()
            assert product.active is True

        mock_message.assert_called_once()

    @patch.object(ProductAdmin, "message_user")
    def test_make_inactive_action(
        self, mock_message, product_admin, admin_request
    ):
        products = [ProductFactory(active=True) for _ in range(2)]
        queryset = Product.objects.filter(id__in=[p.id for p in products])

        product_admin.make_inactive(admin_request, queryset)

        for product in products:
            product.refresh_from_db()
            assert product.active is False

        mock_message.assert_called_once()

    @patch.object(ProductAdmin, "message_user")
    def test_clear_discount_action(
        self, mock_message, product_admin, admin_request
    ):
        products = [ProductFactory(discount_percent=25) for _ in range(2)]
        queryset = Product.objects.filter(id__in=[p.id for p in products])

        product_admin.clear_discount(admin_request, queryset)

        for product in products:
            product.refresh_from_db()
            assert product.discount_percent == 0

        mock_message.assert_called_once()


@pytest.mark.django_db
class TestReviewAdmin:
    def test_review_info_display(self, review_admin):
        review = ProductReviewFactory()

        result = review_admin.review_info(review)

        assert str(review.id) in result
        assert "font-medium" in result

    def test_product_link_display(self, review_admin):
        product = ProductFactory()
        review = ProductReviewFactory(product=product)

        result = review_admin.product_link(review)

        assert (
            product.safe_translation_getter("name", any_language=True) in result
        )

    def test_user_link_display(self, review_admin):
        user = UserAccountFactory()
        review = ProductReviewFactory(user=user)

        result = review_admin.user_link(review)

        assert user.email in result or user.username in result

    def test_rating_display(self, review_admin):
        review = ProductReviewFactory(rate=8)

        result = review_admin.rating_display(review)

        assert "8" in result
        assert "⭐" in result

    def test_status_badge_approved(self, review_admin):
        review = ProductReviewFactory(status="TRUE")

        result = review_admin.status_badge(review)

        assert "✅" in result
        assert "True" in result

    def test_status_badge_pending(self, review_admin):
        review = ProductReviewFactory(status="NEW")

        result = review_admin.status_badge(review)

        assert "🆕" in result
        assert "New" in result

    def test_created_display(self, review_admin):
        review = ProductReviewFactory()

        result = review_admin.created_display(review)

        assert len(result) > 10
        assert "2025" in result or "2024" in result

    @patch.object(ProductReviewAdmin, "message_user")
    def test_approve_reviews_action(
        self, mock_message, review_admin, admin_request
    ):
        reviews = [ProductReviewFactory(status="NEW") for _ in range(3)]
        queryset = ProductReview.objects.filter(id__in=[r.id for r in reviews])

        review_admin.approve_reviews(admin_request, queryset)

        for review in reviews:
            review.refresh_from_db()
            assert review.status == "TRUE"

        mock_message.assert_called_once()

    @patch.object(ProductReviewAdmin, "message_user")
    def test_reject_reviews_action(
        self, mock_message, review_admin, admin_request
    ):
        reviews = [ProductReviewFactory(status="NEW") for _ in range(2)]
        queryset = ProductReview.objects.filter(id__in=[r.id for r in reviews])

        review_admin.reject_reviews(admin_request, queryset)

        for review in reviews:
            review.refresh_from_db()
            assert review.status == "FALSE"

        mock_message.assert_called_once()


@pytest.mark.django_db
class TestFavouriteAdmin:
    def test_user_display(self, favourite_admin):
        user = UserAccountFactory()
        favourite = ProductFavouriteFactory(user=user)

        result = favourite_admin.user_display(favourite)

        assert user.email in result or user.username in result

    def test_product_display(self, favourite_admin):
        product = ProductFactory()
        favourite = ProductFavouriteFactory(product=product)

        result = favourite_admin.product_display(favourite)

        assert (
            product.safe_translation_getter("name", any_language=True) in result
        )

    def test_created_display(self, favourite_admin):
        favourite = ProductFavouriteFactory()

        result = favourite_admin.created_display(favourite)

        assert len(result) > 10
        assert "2025" in result or "2024" in result


@pytest.mark.django_db
class TestProductCategoryImageAdmin:
    def test_image_preview(self, category_image_admin):
        category = ProductCategoryFactory()
        category_image = ProductCategoryImageFactory(category=category)

        result = category_image_admin.image_preview(category_image)

        assert len(result) >= 0

    def test_category_name_display(self, category_image_admin):
        category = ProductCategoryFactory()
        category_image = ProductCategoryImageFactory(category=category)

        result = category_image_admin.category_name(category_image)

        assert (
            category.safe_translation_getter("name", any_language=True)
            in result
        )

    def test_image_type_badge(self, category_image_admin):
        category_image = ProductCategoryImageFactory(image_type="hero")

        result = category_image_admin.image_type_badge(category_image)

        assert "hero" in result.lower()

    def test_status_badge_active(self, category_image_admin):
        category_image = ProductCategoryImageFactory(active=True)

        result = category_image_admin.status_badge(category_image)

        assert "✅" in result
        assert "Active" in result

    def test_status_badge_inactive(self, category_image_admin):
        category_image = ProductCategoryImageFactory(active=False)

        result = category_image_admin.status_badge(category_image)

        assert "❌" in result
        assert "Inactive" in result


@pytest.mark.django_db
class TestProductImageAdmin:
    def test_image_preview(self, product_image_admin):
        product = ProductFactory()
        product_image = ProductImageFactory(product=product)

        result = product_image_admin.image_preview(product_image)

        assert len(result) >= 0

    def test_product_name_display(self, product_image_admin):
        product = ProductFactory()
        product_image = ProductImageFactory(product=product)

        result = product_image_admin.product_name(product_image)

        assert (
            product.safe_translation_getter("name", any_language=True) in result
        )

    def test_main_badge_is_main(self, product_image_admin):
        product_image = ProductImageFactory(is_main=True)

        result = product_image_admin.main_badge(product_image)

        assert "⭐" in result
        assert "Main" in result

    def test_main_badge_not_main(self, product_image_admin):
        product_image = ProductImageFactory(is_main=False)

        result = product_image_admin.main_badge(product_image)

        assert "Gallery" in result


@pytest.mark.django_db
class TestProductAdminIntegration:
    def test_product_admin_configuration(self, product_admin):
        assert product_admin.list_per_page == 25
        assert "product_info" in product_admin.list_display
        assert "pricing_info" in product_admin.list_display
        assert "stock_info" in product_admin.list_display

        action_names = []
        for action in product_admin.actions:
            if hasattr(action, "__name__"):
                action_names.append(action.__name__)
            else:
                action_names.append(str(action))

        assert "make_active" in action_names
        assert "clear_discount" in action_names

    def test_category_admin_configuration(self, category_admin):
        assert category_admin.list_per_page == 25
        assert "category_info" in category_admin.list_display
        assert "category_stats" in category_admin.list_display

    def test_get_queryset_optimization(self, product_admin, admin_request):
        queryset = product_admin.get_queryset(admin_request)

        assert "category" in str(queryset.query)



import uuid
from typing import TYPE_CHECKING

from django.contrib.auth import get_user_model
from django.test import TestCase
from django.utils import timezone

from cart.factories.cart import CartFactory
from cart.factories.item import CartItemFactory
from cart.models import Cart, CartItem
from product.factories.product import ProductFactory
from user.factories.account import UserAccountFactory

if TYPE_CHECKING:
    from product.models.product import Product

User = get_user_model()


class CartModelTestCase(TestCase):
    def setUp(self):
        products = ProductFactory.create_batch(2, num_images=0, num_reviews=0)
        product_1: Product = products[0]
        product_2: Product = products[1]

        self.user = UserAccountFactory(num_addresses=0)
        self.cart = CartFactory(user=self.user, num_cart_items=0)
        self.cart_item_1 = CartItemFactory(
            cart=self.cart, product=product_1, quantity=2
        )
        self.cart_item_2 = CartItemFactory(
            cart=self.cart, product=product_2, quantity=3
        )

    def test_fields(self):
        self.assertEqual(self.cart.user, self.user)
        self.assertEqual(self.cart.last_activity.date(), timezone.now().date())

    def test_str_representation(self):
        expected_str = f"Cart for {self.user} - Items: {self.cart.total_items} - Total: {self.cart.total_price}"
        self.assertEqual(str(self.cart), expected_str)

    def test_get_items(self):
        self.assertEqual(self.cart.get_items().count(), 2)

    def test_total_price(self):
        self.cart_item_1.refresh_from_db()
        self.cart_item_2.refresh_from_db()

        expected_total_price = (
            self.cart_item_1.total_price.amount
            + self.cart_item_2.total_price.amount
        )
        self.assertEqual(self.cart.total_price.amount, expected_total_price)

    def test_total_discount_value(self):
        self.cart_item_1.refresh_from_db()
        self.cart_item_2.refresh_from_db()

        expected_total_discount = (
            self.cart_item_1.total_discount_value.amount
            + self.cart_item_2.total_discount_value.amount
        )
        self.assertEqual(
            self.cart.total_discount_value.amount, expected_total_discount
        )

    def test_total_vat_value(self):
        expected_total_vat = (
            self.cart_item_1.vat_value.amount
            + self.cart_item_2.vat_value.amount
        )
        self.assertEqual(self.cart.total_vat_value.amount, expected_total_vat)

    def test_total_items(self):
        expected_total_items = (
            self.cart_item_1.quantity + self.cart_item_2.quantity
        )
        self.assertEqual(self.cart.total_items, expected_total_items)

    def test_total_items_unique(self):
        self.assertEqual(self.cart.total_items_unique, 2)

    def test_refresh_last_activity(self):
        last_activity_before_refresh = self.cart.last_activity
        self.cart.refresh_last_activity()
        self.assertNotEqual(
            self.cart.last_activity, last_activity_before_refresh
        )


class GuestCartModelTestCase(TestCase):
    def setUp(self):
        products = ProductFactory.create_batch(2, num_images=0, num_reviews=0)
        product_1: Product = products[0]
        product_2: Product = products[1]

        self.session_key = str(uuid.uuid4())
        self.cart = CartFactory(
            user=None, session_key=self.session_key, num_cart_items=0
        )
        self.cart_item_1 = CartItemFactory(
            cart=self.cart, product=product_1, quantity=2
        )
        self.cart_item_2 = CartItemFactory(
            cart=self.cart, product=product_2, quantity=3
        )

    def test_fields(self):
        self.assertIsNone(self.cart.user)
        self.assertEqual(self.cart.session_key, self.session_key)
        self.assertEqual(self.cart.last_activity.date(), timezone.now().date())

    def test_str_representation(self):
        expected_str = f"Guest Cart ({self.session_key[:8]}...) - Items: {self.cart.total_items} - Total: {self.cart.total_price}"
        self.assertEqual(str(self.cart), expected_str)

    def test_get_items(self):
        self.assertEqual(self.cart.get_items().count(), 2)

    def test_total_items(self):
        expected_total_items = (
            self.cart_item_1.quantity + self.cart_item_2.quantity
        )
        self.assertEqual(self.cart.total_items, expected_total_items)

    def test_total_items_unique(self):
        self.assertEqual(self.cart.total_items_unique, 2)

    def test_create_guest_cart(self):
        session_key = str(uuid.uuid4())
        cart = Cart.objects.create(session_key=session_key)
        self.assertIsNotNone(cart)
        self.assertIsNone(cart.user)
        self.assertEqual(cart.session_key, session_key)

    def test_add_item_to_guest_cart(self):
        session_key = str(uuid.uuid4())
        cart = Cart.objects.create(session_key=session_key)
        product = ProductFactory(num_images=0, num_reviews=0)
        cart_item = CartItem.objects.create(
            cart=cart, product=product, quantity=3
        )
        self.assertEqual(cart_item.quantity, 3)
        self.assertEqual(cart_item.product, product)
        self.assertIsNone(cart_item.cart.user)
        self.assertEqual(cart_item.cart.session_key, session_key)



from .account import UserAccount
from .address import UserAddress
from .subscription import SubscriptionTopic, UserSubscription

__all__ = [
    "SubscriptionTopic",
    "UserAccount",
    "UserAddress",
    "UserSubscription",
]



from django.conf import settings
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from country.factories import CountryFactory
from region.factories import RegionFactory
from region.models import Region
from region.serializers import RegionDetailSerializer

languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]
default_language = settings.PARLER_DEFAULT_LANGUAGE_CODE


class RegionViewSetTestCase(APITestCase):
    def setUp(self):
        self.country = CountryFactory(num_regions=0)
        self.region = RegionFactory(
            alpha="GRC",
            country=self.country,
        )

    def get_region_detail_url(self, pk):
        return reverse("region-detail", args=[pk])

    def get_region_list_url(self):
        return reverse("region-list")

    def test_list(self):
        url = self.get_region_list_url()
        response = self.client.get(url)
        regions = Region.objects.all()
        serializer = RegionDetailSerializer(regions, many=True)

        self.assertEqual(response.data["results"], serializer.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_create_valid(self):
        payload = {
            "alpha": "GRD",
            "country": self.region.country.pk,
            "translations": {},
        }
        for language in languages:
            language_code = language[0]
            language_name = language[1]

            translation_payload = {
                "name": f"New Region name in {language_name}",
            }

            payload["translations"][language_code] = translation_payload

        url = self.get_region_list_url()
        response = self.client.post(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_invalid(self):
        payload = {
            "alpha": "invalid_alpha",
            "country": "invalid_country",
            "translations": {
                "invalid_lang_code": {
                    "name": "Translation for invalid language code",
                }
            },
        }

        url = self.get_region_list_url()
        response = self.client.post(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retrieve_valid(self):
        url = self.get_region_detail_url(self.region.alpha)
        response = self.client.get(url)
        region = Region.objects.get(alpha=self.region.alpha)
        serializer = RegionDetailSerializer(region)

        self.assertEqual(response.data, serializer.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_retrieve_invalid(self):
        url = self.get_region_detail_url(999)
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_valid(self):
        payload = {
            "alpha": "GRC",
            "country": self.region.country.pk,
            "translations": {},
        }
        for language in languages:
            language_code = language[0]
            language_name = language[1]

            translation_payload = {
                "name": f"Updated Region name in {language_name}",
            }

            payload["translations"][language_code] = translation_payload

        url = self.get_region_detail_url(self.region.pk)
        response = self.client.put(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_invalid(self):
        payload = {
            "alpha": "invalid_alpha",
            "country": "invalid_country",
            "translations": {
                "invalid_lang_code": {
                    "name": "Translation for invalid language code",
                }
            },
        }

        url = self.get_region_detail_url(self.region.pk)
        response = self.client.put(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_partial_update_valid(self):
        payload = {
            "translations": {
                default_language: {
                    "name": f"Updated Region {default_language}",
                }
            },
        }

        url = self.get_region_detail_url(self.region.pk)
        response = self.client.patch(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_partial_update_invalid(self):
        payload = {
            "alpha": "invalid_alpha",
            "translations": {
                "invalid_lang_code": {
                    "name": "Translation for invalid language code",
                }
            },
        }

        url = self.get_region_detail_url(self.region.pk)
        response = self.client.patch(url, data=payload, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_destroy_valid(self):
        url = self.get_region_detail_url(self.region.pk)
        response = self.client.delete(url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_destroy_invalid(self):
        url = self.get_region_detail_url(999)
        response = self.client.delete(url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



# Generated by Django 5.0.4 on 2024-04-11 11:43
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "cart",
            "0005_cart_cart_created_at_idx_cart_cart_updated_at_idx_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="cartitem",
            options={
                "ordering": ["-created_at"],
                "verbose_name": "Cart Item",
                "verbose_name_plural": "Cart Items",
            },
        ),
    ]
